"""Tests for gdb_convert_to_indexed and gdb_create_linked."""

import contextlib
from pathlib import Path

import pytest

import pymisha as pm


def _write_per_chrom_db(root, chrom_rows, seq_name_map=None):
    seq_dir = root / "seq"
    tracks_dir = root / "tracks"
    pssms_dir = root / "pssms"
    seq_dir.mkdir(parents=True)
    tracks_dir.mkdir()
    pssms_dir.mkdir()

    with open(root / "chrom_sizes.txt", "w", encoding="utf-8") as fh:
        for chrom, seq in chrom_rows:
            fh.write(f"{chrom}\t{len(seq)}\n")
            seq_name = seq_name_map(chrom) if seq_name_map else chrom
            (seq_dir / f"{seq_name}.seq").write_bytes(seq.encode("ascii"))


@pytest.fixture
def restore_db():
    old_root = pm._shared._GROOT
    old_user = pm._shared._UROOT
    yield
    if old_root is None:
        pm.gdb_unload()
    else:
        pm.gdb_init(old_root, old_user)


def test_gdb_convert_to_indexed_preserves_order(tmp_path, restore_db):
    root = tmp_path / "db"
    chrom_rows = [
        ("chr15", "A" * 15),
        ("chr10", "C" * 10),
        ("chr1", "T" * 20),
    ]
    _write_per_chrom_db(root, chrom_rows)

    pm.gdb_convert_to_indexed(groot=str(root), force=True, validate=True)

    assert (root / "seq" / "genome.idx").is_file()
    assert (root / "seq" / "genome.seq").is_file()
    assert (root / "seq" / "genome.seq").read_bytes() == (
        b"A" * 15 + b"C" * 10 + b"T" * 20
    )

    rows = (root / "chrom_sizes.txt").read_text(encoding="utf-8").strip().splitlines()
    names = [row.split("\t")[0] for row in rows]
    assert names == [c for c, _ in chrom_rows]

    info = pm.gdb_info(str(root))
    assert info["is_db"] is True
    assert info["format"] == "indexed"


def test_gdb_convert_to_indexed_resolves_chr_prefix_mismatch(tmp_path, restore_db):
    root = tmp_path / "db"
    chrom_rows = [
        ("15", "A" * 15),
        ("10", "C" * 10),
        ("1", "T" * 20),
    ]
    _write_per_chrom_db(root, chrom_rows, seq_name_map=lambda chrom: f"chr{chrom}")

    pm.gdb_convert_to_indexed(groot=str(root), validate=True)

    rows = (root / "chrom_sizes.txt").read_text(encoding="utf-8").strip().splitlines()
    names = [row.split("\t")[0] for row in rows]
    sizes = [int(row.split("\t")[1]) for row in rows]
    assert names == ["chr15", "chr10", "chr1"]
    assert sizes == [15, 10, 20]


def test_gdb_convert_to_indexed_remove_old_files(tmp_path, restore_db):
    root = tmp_path / "db"
    chrom_rows = [("chr1", "ACTGACTG")]
    _write_per_chrom_db(root, chrom_rows)

    old_seq = root / "seq" / "chr1.seq"
    assert old_seq.exists()

    pm.gdb_convert_to_indexed(groot=str(root), remove_old_files=True, validate=False)

    assert (root / "seq" / "genome.idx").is_file()
    assert (root / "seq" / "genome.seq").is_file()
    assert not old_seq.exists()


def test_gdb_convert_to_indexed_converts_tracks_and_intervals(tmp_path, restore_db):
    root = tmp_path / "db"
    chrom_rows = [("chr1", "A" * 100)]
    _write_per_chrom_db(root, chrom_rows)

    pm.gdb_init(str(root))
    intervals = pm.gintervals(["chr1"], [0], [10])
    pm.gtrack_create_sparse("tmp_sparse", "tmp", intervals, [1.0])

    intervals_dir = root / "tracks" / "tmp_set.interv"
    intervals_dir.mkdir()
    (intervals_dir / "chr1").write_bytes(b"\x01\x02\x03\x04")

    pm.gdb_convert_to_indexed(
        groot=str(root),
        validate=False,
        convert_tracks=True,
        convert_intervals=True,
        remove_old_files=True,
    )

    track_dir = root / "tracks" / "tmp_sparse.track"
    assert (track_dir / "track.idx").is_file()
    assert (track_dir / "track.dat").is_file()

    assert (intervals_dir / "intervals.idx").is_file()
    assert (intervals_dir / "intervals.dat").is_file()
    assert not (intervals_dir / "chr1").exists()


def test_gdb_convert_to_indexed_reverse_chr_prefix(tmp_path, restore_db):
    """Port of: gdb.convert_to_indexed preserves order when removing chr prefix.

    chrom_sizes.txt has 'chr' prefix, but .seq files do NOT.
    """
    root = tmp_path / "db"
    chrom_rows = [
        ("chr15", "A" * 15),
        ("chr10", "C" * 10),
        ("chr1", "T" * 20),
    ]
    seq_dir = root / "seq"
    tracks_dir = root / "tracks"
    pssms_dir = root / "pssms"
    seq_dir.mkdir(parents=True)
    tracks_dir.mkdir()
    pssms_dir.mkdir()

    with open(root / "chrom_sizes.txt", "w", encoding="utf-8") as fh:
        for chrom, seq in chrom_rows:
            fh.write(f"{chrom}\t{len(seq)}\n")
            # Write files WITHOUT chr prefix (strip it)
            name_no_prefix = chrom.replace("chr", "")
            (seq_dir / f"{name_no_prefix}.seq").write_bytes(seq.encode("ascii"))

    pm.gdb_convert_to_indexed(groot=str(root), force=True, validate=False)

    assert (root / "seq" / "genome.idx").is_file()
    assert (root / "seq" / "genome.seq").is_file()

    rows = (root / "chrom_sizes.txt").read_text(encoding="utf-8").strip().splitlines()
    names = [row.split("\t")[0] for row in rows]
    sizes = [int(row.split("\t")[1]) for row in rows]
    # Order should be preserved
    assert sizes == [15, 10, 20]
    # Names should still be recognizable (either with or without chr prefix)
    assert len(names) == 3


def test_gdb_create_linked_creates_symlinked_db(tmp_path):
    parent = tmp_path / "parent"
    _write_per_chrom_db(parent, [("chr1", "ACGT")])
    child = tmp_path / "linked"

    assert pm.gdb_create_linked(str(child), str(parent)) is True

    assert (child / "tracks").is_dir()
    assert (child / "chrom_sizes.txt").is_symlink()
    assert (child / "seq").is_symlink()
    assert (child / "chrom_sizes.txt").resolve() == (parent / "chrom_sizes.txt").resolve()
    assert (child / "seq").resolve() == (parent / "seq").resolve()


def test_gdb_create_linked_validation(tmp_path):
    parent = tmp_path / "parent"
    parent.mkdir()
    (parent / "chrom_sizes.txt").write_text("chr1\t10\n", encoding="utf-8")

    with pytest.raises(FileNotFoundError, match="seq"):
        pm.gdb_create_linked(str(tmp_path / "linked"), str(parent))

    (parent / "seq").mkdir()
    linked = tmp_path / "linked"
    linked.mkdir()
    with pytest.raises(FileExistsError, match="already exists"):
        pm.gdb_create_linked(str(linked), str(parent))


# ===========================================================================
# Track format conversion tests
# (ported from R test-gtrack-format-conversion.R)
# ===========================================================================

TEST_DB = Path(__file__).resolve().parent / "testdb" / "trackdb" / "test"


@pytest.fixture
def restore_test_db():
    """Ensure the test DB is re-initialized after track conversion tests."""
    yield
    pm.gdb_init(str(TEST_DB))


def _cleanup_track(name):
    """Remove a track if it exists, then reload the DB."""
    with contextlib.suppress(Exception):
        pm.gtrack_rm(name, force=True)
    pm._pymisha.pm_dbreload()


def _vals_equal(a, b, tol=1e-6):
    """Compare two float arrays, treating NaN == NaN as True."""
    import numpy as np
    a = np.asarray(a, dtype=float)
    b = np.asarray(b, dtype=float)
    nan_a = np.isnan(a)
    nan_b = np.isnan(b)
    if not (nan_a == nan_b).all():
        return False
    mask = ~nan_a
    if not mask.any():
        return True
    return np.allclose(a[mask], b[mask], atol=tol, rtol=tol)


class TestTrackFormatConversion:
    """Tests for gtrack_convert_to_indexed on the live test DB.

    Ported from R test-gtrack-format-conversion.R.
    """

    def test_gtrack_info_reports_format_per_chromosome(self):
        """gtrack_info reports 'per-chromosome' for standard test tracks.

        Ported from R line 19-32.
        """
        info_dense = pm.gtrack_info("dense_track")
        assert "format" in info_dense
        assert info_dense["format"] == "per-chromosome"

        info_sparse = pm.gtrack_info("sparse_track")
        assert info_sparse["format"] == "per-chromosome"

    def test_convert_nonexistent_track_raises(self):
        """gtrack_convert_to_indexed fails for non-existent track.

        Ported from R line 34-39.
        """
        with pytest.raises(Exception, match="does not exist|not found|Track.*nonexistent"):
            pm.gtrack_convert_to_indexed("nonexistent_track_xyz")

    def test_convert_dense_track(self, restore_test_db):
        """gtrack_convert_to_indexed works for dense tracks.

        Ported from R line 48-60.
        Creates a copy of dense_track, converts it to indexed format,
        then verifies data integrity.
        """
        track = "temp_converted_dense"
        try:
            pm.gtrack_create(track, "", "dense_track")
            assert pm.gtrack_info(track)["format"] == "per-chromosome"

            pm.gtrack_convert_to_indexed(track)
            assert pm.gtrack_info(track)["format"] == "indexed"

            # Verify data is intact
            intervals = pm.gintervals(["1"], 0, 1000)
            r = pm.gextract(track, intervals)
            r_orig = pm.gextract("dense_track", intervals)
            assert _vals_equal(r[track].values, r_orig["dense_track"].values)
        finally:
            _cleanup_track(track)

    def test_convert_sparse_track(self, restore_test_db):
        """gtrack_convert_to_indexed works for sparse tracks.

        Ported from R line 62-74.
        """
        track = "temp_converted_sparse"
        try:
            # Copy sparse track by extracting and re-creating
            intervals = pm.gintervals(["1", "2"])
            orig = pm.gextract("sparse_track", intervals)
            track_intervals = orig[["chrom", "start", "end"]].copy()
            values = orig["sparse_track"].tolist()
            pm.gtrack_create_sparse(track, "", track_intervals, values)
            assert pm.gtrack_info(track)["format"] == "per-chromosome"

            pm.gtrack_convert_to_indexed(track)
            assert pm.gtrack_info(track)["format"] == "indexed"

            # Verify data is intact
            r = pm.gextract(track, intervals)
            assert len(r) == len(orig)
        finally:
            _cleanup_track(track)

    def test_track_expressions_with_converted_dense(self, restore_test_db):
        """Track expressions work with converted dense tracks.

        Ported from R line 94-111.
        Tests arithmetic expressions on converted tracks.
        """
        track = "temp_converted_expr"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            intervals = pm.gintervals(["1"], 0, 5000)

            # Simple addition
            r1 = pm.gextract(f"{track} + 1", intervals)
            r_orig = pm.gextract("dense_track + 1", intervals)
            assert _vals_equal(r1.iloc[:, 3].values, r_orig.iloc[:, 3].values)

            # Multiplication
            r2 = pm.gextract(f"{track} * 2", intervals)
            r_orig2 = pm.gextract("dense_track * 2", intervals)
            assert _vals_equal(r2.iloc[:, 3].values, r_orig2.iloc[:, 3].values)

            # Complex expression across multiple chroms
            intervals_mc = pm.gintervals(["1", "2"], 0, 10000)
            r3 = pm.gextract(f"2 * {track} + 0.5", intervals_mc)
            r_orig3 = pm.gextract("2 * dense_track + 0.5", intervals_mc)
            assert _vals_equal(r3.iloc[:, 3].values, r_orig3.iloc[:, 3].values)
        finally:
            _cleanup_track(track)

    def test_track_expressions_with_converted_sparse(self, restore_test_db):
        """Track expressions work with converted sparse tracks.

        Ported from R line 113-122.
        """
        track = "temp_converted_sparse_expr"
        try:
            # Copy sparse track by extracting and re-creating
            intervals = pm.gintervals(["1", "2"])
            orig = pm.gextract("sparse_track", intervals)
            track_intervals = orig[["chrom", "start", "end"]].copy()
            values = orig["sparse_track"].tolist()
            pm.gtrack_create_sparse(track, "", track_intervals, values)
            pm.gtrack_convert_to_indexed(track)

            # Extract with the same intervals (no explicit iterator needed)
            r1 = pm.gextract(f"{track} * 3", intervals)
            r_orig = pm.gextract("sparse_track * 3", intervals)
            assert len(r1) == len(r_orig)
            assert _vals_equal(r1.iloc[:, 3].values, r_orig.iloc[:, 3].values)
        finally:
            _cleanup_track(track)

    def test_mixed_expressions_converted_and_normal(self, restore_test_db):
        """Mixed expressions with converted and non-converted tracks.

        Ported from R line 124-137.
        """
        track = "temp_converted_mixed"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            intervals = pm.gintervals(["1"], 0, 5000)
            r1 = pm.gextract(f"{track} + dense_track", intervals)
            r_orig = pm.gextract("dense_track + dense_track", intervals)
            assert _vals_equal(r1.iloc[:, 3].values, r_orig.iloc[:, 3].values)

            intervals_mc = pm.gintervals(["1", "2"], 0, 10000)
            r2 = pm.gextract(f"{track} * dense_track + 0.1", intervals_mc)
            r_orig2 = pm.gextract("dense_track * dense_track + 0.1", intervals_mc)
            assert _vals_equal(r2.iloc[:, 3].values, r_orig2.iloc[:, 3].values)
        finally:
            _cleanup_track(track)

    def test_gtrack_create_from_converted_track(self, restore_test_db):
        """gtrack_create with expression using converted track.

        Ported from R line 139-151.
        """
        src_track = "temp_converted_source"
        new_track = "temp_created_from_converted"
        try:
            pm.gtrack_create(src_track, "", "dense_track")
            pm.gtrack_convert_to_indexed(src_track)

            pm.gtrack_create(new_track, "", f"{src_track} * 2 + 5")
            intervals = pm.gintervals(["1"], 0, 5000)
            r1 = pm.gextract(new_track, intervals)
            r_orig = pm.gextract("dense_track * 2 + 5", intervals)
            assert _vals_equal(r1.iloc[:, 3].values, r_orig.iloc[:, 3].values)
        finally:
            _cleanup_track(new_track)
            _cleanup_track(src_track)

    def test_vtrack_on_converted_dense(self, restore_test_db):
        """Vtrack based on converted dense track works.

        Ported from R line 157-169.
        Compare a vtrack on the converted track against a vtrack on the
        original track (both using func='avg'), so the aggregation is
        identical.
        """
        track = "temp_converted_vtrack"
        vt_conv = "v_converted"
        vt_orig = "v_orig"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            pm.gvtrack_create(vt_conv, track, func="avg")
            pm.gvtrack_create(vt_orig, "dense_track", func="avg")
            intervals = pm.gintervals(["1", "2"])
            r = pm.gextract(vt_conv, intervals, iterator=100)
            r_orig = pm.gextract(vt_orig, intervals, iterator=100)
            assert len(r) == len(r_orig)
            assert _vals_equal(r[vt_conv].values, r_orig[vt_orig].values)
        finally:
            for vt in (vt_conv, vt_orig):
                with contextlib.suppress(Exception):
                    pm.gvtrack_rm(vt)
            _cleanup_track(track)

    def test_vtrack_avg_on_converted_track(self, restore_test_db):
        """Vtrack with avg function on converted track.

        Ported from R line 171-185.
        """
        track = "temp_converted_vtrack_avg"
        vt = "v_converted_avg"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            pm.gvtrack_create(vt, track, func="avg")
            intervals = pm.gintervals(["1", "2"])
            r = pm.gextract(vt, intervals, iterator=500)
            r_orig = pm.gextract("dense_track", intervals, iterator=500)
            assert len(r) == len(r_orig)
        finally:
            with contextlib.suppress(Exception):
                pm.gvtrack_rm(vt)
            _cleanup_track(track)

    def test_gtrack_modify_on_converted_track(self, restore_test_db):
        """gtrack_modify works on converted tracks.

        Ported from R line 248-261.
        """
        track = "temp_converted_modify"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            intervals = pm.gintervals(["1"], 1000, 5000)
            before = pm.gextract(track, intervals)

            pm.gtrack_modify(track, f"{track} * 2", intervals)

            after = pm.gextract(track, intervals)
            import numpy as np
            np.testing.assert_allclose(
                after[track].values, before[track].values * 2
            )
        finally:
            _cleanup_track(track)

    def test_numeric_iterator_with_converted_track(self, restore_test_db):
        """Numeric iterator works with converted track.

        Ported from R line 298-308.
        """
        track = "temp_converted_iter"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            intervals = pm.gintervals(["1"], 0, 10000)
            r = pm.gextract(track, intervals, iterator=500)
            r_orig = pm.gextract("dense_track", intervals, iterator=500)
            assert len(r) == len(r_orig)
            assert _vals_equal(r.iloc[:, 3].values, r_orig.iloc[:, 3].values)
        finally:
            _cleanup_track(track)

    def test_gscreen_with_converted_track(self, restore_test_db):
        """gscreen works with converted tracks.

        Ported from R line 350-359.
        """
        track = "temp_converted_screen"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            intervals = pm.gintervals(["1", "2"])
            r = pm.gscreen(f"{track} > 0.2", intervals)
            r_orig = pm.gscreen("dense_track > 0.2", intervals)
            if r is None or r_orig is None:
                assert r is None and r_orig is None
            else:
                assert len(r) == len(r_orig)
        finally:
            _cleanup_track(track)

    def test_converted_track_with_allgenome(self, restore_test_db):
        """Converted track with ALLGENOME extraction.

        Ported from R line 418-428.
        """
        track = "temp_converted_allgenome"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            r = pm.gextract(track, pm.gintervals_all())
            r_orig = pm.gextract("dense_track", pm.gintervals_all())
            assert len(r) == len(r_orig)
        finally:
            _cleanup_track(track)

    def test_multiple_operations_on_converted_track(self, restore_test_db):
        """Multiple operations on same converted track.

        Ported from R line 389-416.
        Tests extract, vtrack, and modify in sequence on one converted track.
        """
        track = "temp_converted_multi_ops"
        vt = "v_multi"
        try:
            pm.gtrack_create(track, "", "dense_track")
            pm.gtrack_convert_to_indexed(track)

            # Extract with expression
            intervals_small = pm.gintervals(["1"], 0, 5000)
            r1 = pm.gextract(f"{track} * 2", intervals_small)
            assert len(r1) > 0

            # Create vtrack and use expression
            pm.gvtrack_create(vt, track)
            r2 = pm.gextract(f"{vt} + 0.5", intervals_small, iterator=100)
            assert len(r2) > 0

            # Modify in-place
            intervs = pm.gintervals(["1"], 1000, 3000)
            pm.gtrack_modify(track, f"{track} * 1.5", intervs)
            r3 = pm.gextract(track, intervs)
            assert len(r3) > 0

            # Extract again after modification
            r4 = pm.gextract(track, intervals_small)
            assert len(r4) > 0
        finally:
            with contextlib.suppress(Exception):
                pm.gvtrack_rm(vt)
            _cleanup_track(track)



"""Tests for streaming gdist/gsummary/gquantiles with virtual track expressions.

These tests verify that summary/distribution functions work correctly when
expressions contain virtual tracks, using a chunked streaming approach that
keeps memory bounded regardless of genome size.
"""

import numpy as np
import pandas as pd
import pytest

import pymisha as pm


@pytest.fixture(autouse=True)
def setup_db():
    """Initialize example database and clear vtracks for each test."""
    pm.gdb_init_examples()
    pm.gvtrack_clear()
    yield
    pm.gvtrack_clear()


# ---------------------------------------------------------------------------
# gdist with virtual tracks
# ---------------------------------------------------------------------------

class TestGdistWithVtracks:
    """Test gdist produces correct results when expressions use vtracks."""

    def test_gdist_vtrack_matches_manual_binning(self):
        """gdist with a vtrack expression should match manual extract+bin."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks, intervals=intervals, iterator=100)

        # Manual: extract then bin
        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        expected = _manual_bin_count(vals, breaks, include_lowest=False)

        np.testing.assert_array_equal(result, expected)

    def test_gdist_vtrack_include_lowest(self):
        """gdist with vtrack and include_lowest=True matches manual binning."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks, intervals=intervals,
                          iterator=100, include_lowest=True)

        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        expected = _manual_bin_count(vals, breaks, include_lowest=True)

        np.testing.assert_array_equal(result, expected)

    def test_gdist_vtrack_2d(self):
        """gdist with two vtrack expressions (2D distribution)."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        pm.gvtrack_create("vt_max", "dense_track", func="max")
        intervals = pm.gintervals("1", 0, 10000)
        breaks1 = [0, 0.2, 0.5, 1.0]
        breaks2 = [0, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks1, "vt_max", breaks2,
                          intervals=intervals, iterator=100)

        assert result.shape == (3, 2)
        assert np.all(result >= 0)

        # Verify against manual computation
        vals1 = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        vals2 = _extract_vtrack_values("vt_max", intervals, iterator=100)
        expected = _manual_bin_count_2d(vals1, breaks1, vals2, breaks2,
                                       include_lowest=False)
        np.testing.assert_array_equal(result, expected)

    def test_gdist_mixed_vtrack_physical(self):
        """gdist with one vtrack and one physical track expression."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks1 = [0, 0.2, 0.5, 1.0]
        breaks2 = [0, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks1, "dense_track", breaks2,
                          intervals=intervals, iterator=100)

        assert result.shape == (3, 2)
        assert np.all(result >= 0)
        assert result.sum() > 0

    def test_gdist_vtrack_dataframe(self):
        """gdist with vtrack returns correct DataFrame when dataframe=True."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.2, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks, intervals=intervals,
                          iterator=100, dataframe=True)

        assert isinstance(result, pd.DataFrame)
        assert "n" in result.columns
        assert len(result) == 3  # 3 bins

    def test_gdist_vtrack_empty_intervals(self):
        """gdist with vtrack and empty intervals returns zeros."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        empty = pd.DataFrame(columns=["chrom", "start", "end"])
        breaks = [0, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks, intervals=empty, iterator=100)
        np.testing.assert_array_equal(result, [0, 0])

    def test_gdist_vtrack_with_shift(self):
        """gdist with a shifted vtrack still works correctly."""
        pm.gvtrack_create("vt_shifted", "dense_track", func="avg",
                          sshift=-50, eshift=50)
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.2, 0.5, 1.0]

        result = pm.gdist("vt_shifted", breaks, intervals=intervals,
                          iterator=100)

        # Verify against manual
        vals = _extract_vtrack_values("vt_shifted", intervals, iterator=100)
        expected = _manual_bin_count(vals, breaks, include_lowest=False)
        np.testing.assert_array_equal(result, expected)

    def test_gdist_vtrack_expression_arithmetic(self):
        """gdist with vtrack used in an arithmetic expression."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.4, 1.0, 2.0]

        result = pm.gdist("vt_avg * 2", breaks, intervals=intervals,
                          iterator=100)

        assert result is not None
        assert len(result) == 3
        # Total should equal the number of values within the break range
        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        doubled = vals * 2
        expected = _manual_bin_count(doubled, breaks, include_lowest=False)
        np.testing.assert_array_equal(result, expected)

    def test_gdist_vtrack_consistent_with_cpp_binning_semantics(self):
        """Verify that vtrack gdist uses the same binning semantics as C++ path.

        Bins are (breaks[i], breaks[i+1]] — open on left, closed on right.
        Values exactly at breaks[0] are excluded unless include_lowest=True.
        """
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        result_default = pm.gdist("vt_avg", breaks, intervals=intervals,
                                  iterator=100)
        result_include = pm.gdist("vt_avg", breaks, intervals=intervals,
                                  iterator=100, include_lowest=True)

        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        n_at_zero = int(np.count_nonzero(vals == 0.0))

        # include_lowest should capture n_at_zero additional values
        assert result_include.sum() - result_default.sum() == n_at_zero


# ---------------------------------------------------------------------------
# gsummary with virtual tracks (streaming)
# ---------------------------------------------------------------------------

class TestGsummaryWithVtracks:
    """Test gsummary with vtrack expressions."""

    def test_gsummary_vtrack_matches_manual(self):
        """gsummary with vtrack should match manual computation."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)

        result = pm.gsummary("vt_avg", intervals=intervals, iterator=100)

        # Manual computation
        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        valid = vals[~np.isnan(vals)]

        assert result["Total intervals"] == len(vals)
        assert result["NaN intervals"] == np.count_nonzero(np.isnan(vals))
        if len(valid) > 0:
            np.testing.assert_allclose(result["Min"], np.min(valid), rtol=1e-10)
            np.testing.assert_allclose(result["Max"], np.max(valid), rtol=1e-10)
            np.testing.assert_allclose(result["Sum"], np.sum(valid), rtol=1e-10)
            np.testing.assert_allclose(result["Mean"], np.mean(valid), rtol=1e-10)


# ---------------------------------------------------------------------------
# gquantiles with virtual tracks
# ---------------------------------------------------------------------------

class TestGquantilesWithVtracks:
    """Test gquantiles with vtrack expressions."""

    def test_gquantiles_vtrack_matches_manual(self):
        """gquantiles with vtrack should match manual quantile computation."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        pcts = [0.25, 0.5, 0.75]

        result = pm.gquantiles("vt_avg", pcts, intervals=intervals,
                               iterator=100)

        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        valid = vals[~np.isnan(vals)]
        expected = np.quantile(valid, pcts) if len(valid) > 0 else [np.nan] * 3

        np.testing.assert_allclose(result.values, expected, rtol=1e-10)

    def test_gquantiles_vtrack_uses_streaming_path(self, monkeypatch):
        """gquantiles vtrack path should not materialize via _extract_expr_values."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)

        import pymisha.summary as summary_mod

        def _unexpected_extract(*args, **kwargs):
            raise AssertionError("vtrack gquantiles should not call _extract_expr_values")

        monkeypatch.setattr(summary_mod, "_extract_expr_values", _unexpected_extract)
        result = pm.gquantiles("vt_avg", [0.5], intervals=intervals, iterator=100)
        assert len(result) == 1

    def test_gquantiles_vtrack_warns_when_sampling_is_capped(self):
        """gquantiles should warn when vtrack data exceeds max_data_size."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        starts = np.arange(0, 50000, 100, dtype=int)
        ends = starts + 100
        intervals = pm.gintervals(["1"] * len(starts), starts, ends)

        old_max_data_size = pm.CONFIG.get("max_data_size")
        pm.CONFIG["max_data_size"] = 100
        try:
            with pytest.warns(RuntimeWarning, match="Data size exceeds the limit; quantiles are approximate"):
                result = pm.gquantiles("vt_avg", [0.25, 0.5, 0.75], intervals=intervals)
            assert len(result) == 3
        finally:
            pm.CONFIG["max_data_size"] = old_max_data_size


# ---------------------------------------------------------------------------
# Binning bug regression test
# ---------------------------------------------------------------------------

class TestGdistBinningSemantics:
    """Regression tests for binning edge cases (values at breaks[0])."""

    def test_values_at_breaks0_excluded_without_include_lowest(self):
        """Values exactly at breaks[0] must be excluded when include_lowest=False.

        This is a regression test for a bug where searchsorted mapped values
        at breaks[0] into bin 0 even without include_lowest.
        """
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        int(np.count_nonzero(vals == breaks[0]))

        result = pm.gdist("vt_avg", breaks, intervals=intervals, iterator=100)

        # Total should equal len(vals) - n_at_break0 - n_nan - n_out_of_range
        int(np.count_nonzero(np.isnan(vals)))
        n_valid = sum(1 for v in vals if not np.isnan(v) and v > breaks[0] and v <= breaks[-1])
        assert result.sum() == n_valid

    def test_values_at_breaks0_included_with_include_lowest(self):
        """Values exactly at breaks[0] must be in bin 0 when include_lowest=True."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)

        result = pm.gdist("vt_avg", breaks, intervals=intervals,
                          iterator=100, include_lowest=True)

        int(np.count_nonzero(np.isnan(vals)))
        n_valid = sum(1 for v in vals if not np.isnan(v) and v >= breaks[0] and v <= breaks[-1])
        assert result.sum() == n_valid

    def test_cpp_path_binning_matches_semantics(self):
        """C++ path should also exclude values at breaks[0] without include_lowest.

        This test uses a physical track (C++ path) to verify.
        """
        intervals = pm.gintervals("1", 0, 10000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        result_default = pm.gdist("dense_track", breaks,
                                  intervals=intervals, iterator=100)
        result_include = pm.gdist("dense_track", breaks,
                                  intervals=intervals, iterator=100,
                                  include_lowest=True)

        # include_lowest should capture more values
        assert result_include.sum() >= result_default.sum()


# ---------------------------------------------------------------------------
# Memory-bounded streaming verification
# ---------------------------------------------------------------------------

class TestStreamingBehavior:
    """Verify that vtrack paths use chunked streaming (not full materialization)."""

    def test_gdist_vtrack_larger_scope(self):
        """gdist with vtrack on a larger scope matches manual computation."""
        pm.gvtrack_create("vt_avg", "dense_track", func="avg")
        intervals = pm.gintervals("1", 0, 50000)
        breaks = [0, 0.1, 0.2, 0.5, 1.0]

        result = pm.gdist("vt_avg", breaks, intervals=intervals, iterator=100)

        assert result is not None
        assert len(result) == 4
        assert result.sum() > 0

        # Verify against manual full-extract approach
        vals = _extract_vtrack_values("vt_avg", intervals, iterator=100)
        expected = _manual_bin_count(vals, breaks, include_lowest=False)
        np.testing.assert_array_equal(result, expected)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _extract_vtrack_values(vtrack_name, intervals, iterator=None):
    """Extract values from a vtrack expression via gextract."""
    df = pm.gextract(vtrack_name, intervals, iterator=iterator)
    if df is None or len(df) == 0:
        return np.array([], dtype=float)
    data_cols = [c for c in df.columns
                 if c not in {"chrom", "start", "end", "intervalID"}]
    assert len(data_cols) == 1
    return df[data_cols[0]].to_numpy(dtype=float, copy=False)


def _manual_bin_count(values, breaks, include_lowest=False):
    """Manually bin values and count, matching gdist/C++ BinFinder semantics.

    Bins are (breaks[i], breaks[i+1]] — open on left, closed on right.
    With include_lowest, first bin is [breaks[0], breaks[1]].
    """
    breaks = np.asarray(breaks, dtype=float)
    n_bins = len(breaks) - 1
    counts = np.zeros(n_bins, dtype=int)

    for v in values:
        if np.isnan(v):
            continue
        if v > breaks[-1]:
            continue
        if include_lowest:
            if v < breaks[0]:
                continue
        else:
            if v <= breaks[0]:
                continue
        idx = np.searchsorted(breaks, v, side='right') - 1
        if include_lowest and v == breaks[0]:
            idx = 0
        if 0 <= idx < n_bins:
            counts[idx] += 1

    return counts


def _manual_bin_count_2d(vals1, breaks1, vals2, breaks2, include_lowest=False):
    """Manually compute 2D bin counts."""
    breaks1 = np.asarray(breaks1, dtype=float)
    breaks2 = np.asarray(breaks2, dtype=float)
    n1 = len(breaks1) - 1
    n2 = len(breaks2) - 1
    counts = np.zeros((n1, n2), dtype=int)

    for v1, v2 in zip(vals1, vals2, strict=False):
        if np.isnan(v1) or np.isnan(v2):
            continue

        def _bin_idx(v, brk):
            if v > brk[-1]:
                return -1
            if include_lowest:
                if v < brk[0]:
                    return -1
            else:
                if v <= brk[0]:
                    return -1
            idx = np.searchsorted(brk, v, side='right') - 1
            if include_lowest and v == brk[0]:
                idx = 0
            if 0 <= idx < len(brk) - 1:
                return idx
            return -1

        i1 = _bin_idx(v1, breaks1)
        i2 = _bin_idx(v2, breaks2)
        if i1 >= 0 and i2 >= 0:
            counts[i1, i2] += 1

    return counts



"""Database initialization and example DB helpers."""

import atexit
import contextlib
import copy
import os
import shutil
import struct
import tempfile
from pathlib import Path

import pandas as pd

from . import _shared
from ._shared import CONFIG, _checkroot, _pymisha

_EXAMPLE_TMP_DIRS = []


def _cleanup_example_tmpdirs():
    for tmp in _EXAMPLE_TMP_DIRS:
        with contextlib.suppress(Exception):
            shutil.rmtree(tmp, ignore_errors=True)


atexit.register(_cleanup_example_tmpdirs)


def gdb_init(path: str, userpath: str = None):
    """
    Initialize connection to a misha genomic database.

    Loads the genome database at the given path and makes it available for
    all subsequent genomic operations. Must be called before any other
    pymisha function that accesses track data.

    Parameters
    ----------
    path : str
        Path to the root directory of the genome database.
    userpath : str, optional
        Path to a user-writable database root. New tracks and interval
        sets will be created here. If None, defaults to ``path``.

    Returns
    -------
    None

    See Also
    --------
    gdb_reload : Refresh track lists after external changes.
    gdb_unload : Disconnect from the database and clear all state.
    gdb_info : Return metadata about the database.
    gsetroot : Alternative entry point with directory validation.

    Examples
    --------
    >>> import pymisha as pm
    >>> _ = pm.gdb_init_examples()  # initializes a real test DB
    """
    db_path = Path(path).expanduser()
    if not db_path.exists():
        raise FileNotFoundError(f"Database path does not exist: {path}")
    _shared._GROOT = str(db_path)
    _shared._UROOT = userpath
    _shared._GWD = str(db_path / "tracks")
    _shared._GDATASETS = []
    _shared._VTRACKS = {}

    _pymisha.pm_dbinit(str(db_path), userpath or "", CONFIG)
    _pymisha.pm_dbsetdatasets([])


def gdb_reload():
    """
    Reload the database, refreshing track lists and metadata.

    Re-scans the database root directories for newly created or removed
    tracks and interval sets. Call this after external modifications to
    the database on disk (e.g., tracks created by R misha or another
    process).

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If no database is currently initialized.

    See Also
    --------
    gdb_init : Initialize a database connection.
    gdb_unload : Disconnect from the database entirely.

    Examples
    --------
    >>> import pymisha as pm
    >>> _ = pm.gdb_init_examples()
    >>> pm.gdb_reload()
    """
    _checkroot()
    _pymisha.pm_dbreload()


def gdb_unload():
    """
    Unload the database, clearing all state.

    Disconnects from the currently active genome database and resets all
    internal state including the database root paths, working directory,
    datasets, and virtual tracks. After calling this function, a new
    :func:`gdb_init` call is required before any genomic operations.

    Returns
    -------
    None

    See Also
    --------
    gdb_init : Initialize a new database connection.
    gdb_reload : Refresh without disconnecting.

    Examples
    --------
    >>> import pymisha as pm
    >>> _ = pm.gdb_init_examples()
    >>> pm.gdb_unload()
    """
    _pymisha.pm_dbunload()
    _shared._GROOT = None
    _shared._UROOT = None
    _shared._GWD = None
    _shared._GDATASETS = []
    _shared._VTRACKS = {}


def gdb_examples_path():
    """
    Return the path to the example database if available.

    Checks the following locations in order:
    1) PYMISHA_EXAMPLES_DB environment variable
    2) pymisha/examples/trackdb/test (if packaged)
    3) tests/testdb/trackdb/test (repo checkout)

    Returns
    -------
    str
        Absolute path to the example database root directory.

    Raises
    ------
    FileNotFoundError
        If the example database cannot be located in any of the
        searched locations.

    See Also
    --------
    gdb_init_examples : Initialize the example database.
    gdb_init : Initialize a custom database.

    Examples
    --------
    >>> import pymisha as pm
    >>> path = pm.gdb_examples_path()  # doctest: +ELLIPSIS
    >>> path  # doctest: +SKIP
    """
    env = os.environ.get("PYMISHA_EXAMPLES_DB")
    if env:
        return env

    here = Path(__file__).resolve()
    candidates = [
        here.parent / "examples" / "trackdb" / "test",
        here.parents[1] / "tests" / "testdb" / "trackdb" / "test",
    ]

    for candidate in candidates:
        if candidate.exists():
            return str(candidate)

    raise FileNotFoundError(
        "Example database not found. Set PYMISHA_EXAMPLES_DB to a trackdb/test directory."
    )


def gsetroot(groot, subdir=None, rescan=False, **kwargs):
    """
    Set the database root directory with validation.

    Connects to a genome database after verifying that the directory
    exists and contains the required ``tracks/`` and ``seq/``
    subdirectories. This matches the R ``gsetroot()`` interface and is
    the recommended entry point when working interactively, since it
    provides clear error messages for invalid database paths.

    Parameters
    ----------
    groot : str
        Path to the genome database root directory.
    subdir : str, optional
        Sub-directory within ``tracks/`` to use as working directory
        after initialization.
    dir : str, optional
        Backward-compatible alias for ``subdir``.
    rescan : bool, default False
        If True, force a rescan of the database after initialization.
        Equivalent to calling :func:`gdb_reload` after :func:`gdb_init`.

    Returns
    -------
    None

    Raises
    ------
    FileNotFoundError
        If ``groot`` does not exist, or is missing the required
        ``tracks/`` or ``seq/`` subdirectories.

    See Also
    --------
    gdb_init : Lower-level initializer without directory validation.
    gdb_reload : Refresh track lists without re-initializing.

    Examples
    --------
    >>> import pymisha as pm
    >>> pm.gsetroot(pm.gdb_examples_path())
    """
    if "dir" in kwargs:
        if subdir is not None:
            raise ValueError("Specify only one of 'subdir' or 'dir'")
        subdir = kwargs.pop("dir")
    if kwargs:
        bad = ", ".join(sorted(kwargs))
        raise TypeError(f"Unexpected keyword argument(s): {bad}")

    p = Path(groot)
    if not p.exists():
        raise FileNotFoundError(f"Database directory does not exist: {groot}")
    if not (p / "tracks").exists():
        raise FileNotFoundError(
            f"Database directory '{groot}' does not contain a 'tracks' subdirectory. "
            "This does not appear to be a valid misha database."
        )
    if not (p / "seq").exists():
        raise FileNotFoundError(
            f"Database directory '{groot}' does not contain a 'seq' subdirectory. "
            "This does not appear to be a valid misha database."
        )

    gdb_init(str(p.resolve()))

    if subdir is not None:
        from .gdir import gdir_cd

        gdir_cd(subdir)

    if rescan:
        gdb_reload()


def gdb_init_examples(copy=True):
    """
    Initialize the example database (mirrors R's gdb.init_examples).

    Parameters
    ----------
    copy : bool, default True
        If True, copy the example DB into a temp dir before initializing.
        This avoids mutating the repo data when running examples.

    Returns
    -------
    str
        Path to the initialized example DB.

    See Also
    --------
    gdb_examples_path : Get the path to the example database.
    gdb_init : Initialize a custom database.

    Examples
    --------
    >>> import pymisha as pm
    >>> _ = pm.gdb_init_examples()
    >>> pm.gtrack_ls()  # doctest: +NORMALIZE_WHITESPACE
    ['array_track', 'dense_track', 'rects_track', 'sparse_track', 'subdir.dense_track2']
    """
    src = Path(gdb_examples_path())
    if copy:
        tmpdir = Path(tempfile.mkdtemp(prefix="pymisha-example-"))
        _EXAMPLE_TMP_DIRS.append(tmpdir)
        dst = tmpdir / "trackdb" / "test"
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(src, dst)
        gdb_init(str(dst))
        return str(dst)

    gdb_init(str(src))
    return str(src)


def gdb_info(groot: str = None):
    """
    Return high-level information about a misha database.

    Inspects a genome database directory and returns metadata including
    the storage format, number of chromosomes, total genome size, and a
    table of per-chromosome sizes. Can be used to validate a database
    path without fully initializing a connection.

    Parameters
    ----------
    groot : str, optional
        Path to a database root directory. If ``None``, uses the
        currently initialized database.

    Returns
    -------
    dict
        Dictionary with keys:

        - ``path`` (str) -- Resolved absolute path to the database.
        - ``is_db`` (bool) -- Whether the path is a valid misha database.
        - ``error`` (str) -- Present only when ``is_db`` is False;
          describes why validation failed.
        - ``format`` (str) -- ``"indexed"`` or ``"per-chromosome"``.
          Present only when ``is_db`` is True.
        - ``num_chromosomes`` (int) -- Number of chromosomes.
          Present only when ``is_db`` is True.
        - ``genome_size`` (int) -- Sum of all chromosome sizes.
          Present only when ``is_db`` is True.
        - ``chromosomes`` (pandas.DataFrame) -- Two-column table with
          ``chrom`` and ``size``. Present only when ``is_db`` is True.

    Raises
    ------
    ValueError
        If ``groot`` is None and no database is currently initialized.

    See Also
    --------
    gdb_init : Initialize a database connection.

    Examples
    --------
    >>> import pymisha as pm
    >>> _ = pm.gdb_init_examples()
    >>> info = pm.gdb_info()
    >>> info["num_chromosomes"]
    3
    >>> info["genome_size"]
    1000000
    """
    if groot is None:
        if _shared._GROOT is None:
            raise ValueError(
                "No database is currently active. Call gdb_init() or pass groot."
            )
        groot = _shared._GROOT

    db_path = str(Path(groot).expanduser().resolve(strict=False))

    if not os.path.isdir(db_path):
        return {
            "path": db_path,
            "is_db": False,
            "error": "Directory does not exist",
        }

    chrom_sizes_path = os.path.join(db_path, "chrom_sizes.txt")
    if not os.path.exists(chrom_sizes_path):
        return {
            "path": db_path,
            "is_db": False,
            "error": "Not a misha database (chrom_sizes.txt not found)",
        }

    try:
        chrom_sizes = pd.read_csv(
            chrom_sizes_path,
            sep="\t",
            header=None,
            names=["chrom", "size"],
            dtype={"chrom": str, "size": "int64"},
        )
    except Exception:
        return {
            "path": db_path,
            "is_db": False,
            "error": "Invalid chrom_sizes.txt format",
        }

    if chrom_sizes.empty:
        return {
            "path": db_path,
            "is_db": False,
            "error": "Invalid chrom_sizes.txt format",
        }

    is_indexed = (
        os.path.exists(os.path.join(db_path, "seq", "genome.idx"))
        and os.path.exists(os.path.join(db_path, "seq", "genome.seq"))
    )

    return {
        "path": db_path,
        "is_db": True,
        "format": "indexed" if is_indexed else "per-chromosome",
        "num_chromosomes": int(len(chrom_sizes)),
        "genome_size": int(chrom_sizes["size"].sum()),
        "chromosomes": chrom_sizes,
    }


def _read_chrom_sizes_rows(chrom_sizes_path):
    rows = []
    with open(chrom_sizes_path, encoding="utf-8") as fh:
        for line_num, line in enumerate(fh, start=1):
            line = line.strip()
            if not line:
                continue
            parts = line.split("\t")
            if len(parts) != 2:
                raise ValueError(
                    f"Invalid chrom_sizes.txt format at line {line_num}: expected 2 columns"
                )
            chrom = parts[0].strip()
            if not chrom:
                raise ValueError(
                    f"Invalid chrom_sizes.txt format at line {line_num}: empty chromosome name"
                )
            try:
                size = int(parts[1])
            except ValueError as exc:
                raise ValueError(
                    f"Invalid chrom_sizes.txt format at line {line_num}: invalid size"
                ) from exc
            if size < 0:
                raise ValueError(
                    f"Invalid chrom_sizes.txt format at line {line_num}: size must be non-negative"
                )
            rows.append((chrom, size))
    return rows


def _resolve_seq_file(seq_dir, chrom):
    candidates = [chrom]
    if chrom.startswith("chr"):
        candidates.append(chrom[3:])
    else:
        candidates.append(f"chr{chrom}")

    for name in candidates:
        path = seq_dir / f"{name}.seq"
        if path.exists():
            return path

    tried = ", ".join(f"{candidate}.seq" for candidate in candidates)
    raise FileNotFoundError(
        f"Missing sequence file for chromosome '{chrom}' (tried: {tried})"
    )


def _iter_genome_idx_entries(index_path):
    with open(index_path, "rb") as fh:
        header = fh.read(24)
        if len(header) != 24:
            raise ValueError("Invalid genome.idx header")
        magic, version, num_entries, _checksum = struct.unpack("<8sIIQ", header)
        if magic != b"MISHAIDX":
            raise ValueError("Invalid genome.idx magic")
        if version != 1:
            raise ValueError(f"Unsupported genome.idx version {version}")
        if num_entries > 20000000:
            raise ValueError("Invalid genome.idx entry count")

        entries = []
        for _ in range(num_entries):
            id_and_len = fh.read(6)
            if len(id_and_len) != 6:
                raise ValueError("Truncated genome.idx entry table")
            chrom_id, name_len = struct.unpack("<IH", id_and_len)

            name_bytes = fh.read(name_len)
            if len(name_bytes) != name_len:
                raise ValueError("Truncated genome.idx contig name")

            tail = fh.read(24)
            if len(tail) != 24:
                raise ValueError("Truncated genome.idx entry table")
            offset, length, _reserved = struct.unpack("<QQQ", tail)

            try:
                name = name_bytes.decode("utf-8")
            except UnicodeDecodeError as exc:
                raise ValueError("Invalid UTF-8 contig name in genome.idx") from exc
            entries.append((chrom_id, name, offset, length))

    entries.sort(key=lambda item: item[0])
    return entries


def _stream_wrapped_sequence(out_fh, seq_fh, chrom, length, line_width, chunk_size):
    pending = b""
    remaining = int(length)

    while remaining > 0:
        to_read = min(chunk_size, remaining)
        chunk = seq_fh.read(to_read)
        if len(chunk) != to_read:
            raise OSError(
                f"Failed reading sequence for chromosome {chrom}: "
                f"expected {to_read} bytes, got {len(chunk)}"
            )

        merged = pending + chunk
        full_len = (len(merged) // line_width) * line_width
        if full_len:
            block = merged[:full_len]
            out_fh.write(
                b"\n".join(
                    block[i: i + line_width] for i in range(0, full_len, line_width)
                )
            )
            out_fh.write(b"\n")
        pending = merged[full_len:]
        remaining -= to_read

    if pending:
        out_fh.write(pending)
        out_fh.write(b"\n")


@contextlib.contextmanager
def _temporary_db_root(groot):
    if groot is None:
        _checkroot()
        yield
        return

    root_path = Path(groot).expanduser()
    if not root_path.exists():
        raise FileNotFoundError(f"Database directory does not exist: {groot}")
    target_root = str(root_path.resolve(strict=False))

    old_root = _shared._GROOT
    old_user = _shared._UROOT
    old_gwd = _shared._GWD
    old_datasets = list(_shared._GDATASETS)
    old_vtracks = copy.deepcopy(_shared._VTRACKS)

    old_root_resolved = (
        str(Path(old_root).expanduser().resolve(strict=False))
        if old_root is not None
        else None
    )
    if old_root_resolved == target_root:
        yield
        return

    gdb_init(target_root)
    try:
        yield
    finally:
        if old_root is None:
            gdb_unload()
        else:
            gdb_init(old_root, old_user)
            _shared._GDATASETS = list(old_datasets)
            _shared._pymisha.pm_dbsetdatasets(old_datasets)
            _shared._VTRACKS = old_vtracks
            _shared._GWD = old_gwd


def gdb_export_fasta(
    file=None,
    groot=None,
    line_width=80,
    chunk_size=1000000,
    overwrite=False,
    verbose=False,
):
    """
    Export database genome sequence to a multi-FASTA file.

    Parameters
    ----------
    file : str or os.PathLike
        Output FASTA file path.
    groot : str, optional
        Database root to export. If None, exports the currently active DB.
    line_width : int, default 80
        Number of bases per FASTA line.
    chunk_size : int, default 1000000
        Number of bases to read per I/O chunk.
    overwrite : bool, default False
        If True, replace an existing output file.
    verbose : bool, default False
        If True, print per-chromosome progress.

    Returns
    -------
    str
        Output FASTA path.
    """
    usage = (
        "Usage: gdb_export_fasta(file, groot=None, line_width=80, "
        "chunk_size=1000000, overwrite=False, verbose=False)"
    )
    if file is None or not isinstance(file, (str, os.PathLike)):
        raise ValueError(usage)
    file_str = os.fspath(file)
    if file_str == "":
        raise ValueError(usage)

    try:
        line_width = int(line_width)
    except (TypeError, ValueError) as exc:
        raise ValueError("line_width must be a positive integer") from exc
    if line_width < 1:
        raise ValueError("line_width must be a positive integer")

    try:
        chunk_size = int(chunk_size)
    except (TypeError, ValueError) as exc:
        raise ValueError("chunk_size must be a positive integer") from exc
    if chunk_size < 1:
        raise ValueError("chunk_size must be a positive integer")

    out_path = Path(file_str).expanduser()
    out_dir = out_path.parent
    if not out_dir.exists():
        raise FileNotFoundError(f"Output directory does not exist: {out_dir}")
    if out_path.exists() and not overwrite:
        raise FileExistsError(
            f"Output file already exists: {out_path}. Use overwrite=True to replace it."
        )

    tmp_fd, tmp_name = tempfile.mkstemp(
        prefix="pymisha_export_",
        suffix=".fasta",
        dir=str(out_dir),
    )
    os.close(tmp_fd)
    tmp_path = Path(tmp_name)

    try:
        with _temporary_db_root(groot):
            db_root = Path(_shared._GROOT).expanduser().resolve(strict=False)
            seq_dir = db_root / "seq"
            if not seq_dir.is_dir():
                raise FileNotFoundError(f"seq directory does not exist: {seq_dir}")

            genome_idx = seq_dir / "genome.idx"
            genome_seq = seq_dir / "genome.seq"
            chrom_sizes_path = db_root / "chrom_sizes.txt"
            if not chrom_sizes_path.exists():
                raise FileNotFoundError(
                    f"chrom_sizes.txt not found: {chrom_sizes_path}"
                )

            with open(tmp_path, "wb") as out_fh:
                if genome_idx.exists() and genome_seq.exists():
                    entries = _iter_genome_idx_entries(genome_idx)
                    if not entries:
                        raise ValueError("No chromosomes found in the database")
                    with open(genome_seq, "rb") as seq_fh:
                        for _chrom_id, chrom, offset, length in entries:
                            if verbose:
                                print(f"Exporting {chrom} ({length} bp)")
                            out_fh.write(f">{chrom}\n".encode())
                            seq_fh.seek(offset)
                            _stream_wrapped_sequence(
                                out_fh,
                                seq_fh,
                                chrom,
                                length,
                                line_width,
                                chunk_size,
                            )
                else:
                    chrom_rows = _read_chrom_sizes_rows(chrom_sizes_path)
                    if not chrom_rows:
                        raise ValueError("No chromosomes found in the database")
                    for chrom, length in chrom_rows:
                        seq_path = _resolve_seq_file(seq_dir, chrom)
                        if verbose:
                            print(f"Exporting {chrom} ({length} bp)")
                        out_fh.write(f">{chrom}\n".encode())
                        with open(seq_path, "rb") as seq_fh:
                            _stream_wrapped_sequence(
                                out_fh,
                                seq_fh,
                                chrom,
                                length,
                                line_width,
                                chunk_size,
                            )

        os.replace(tmp_path, out_path)
    except Exception:
        with contextlib.suppress(OSError):
            tmp_path.unlink()
        raise

    return str(out_path)



import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import pymisha as pm

TEST_DB = Path(__file__).resolve().parent / "testdb" / "trackdb" / "test"


def _copy_db(tmp_path: Path) -> Path:
    dst = tmp_path / "trackdb" / "test"
    dst.parent.mkdir(parents=True, exist_ok=True)
    shutil.copytree(TEST_DB, dst)
    return dst


def test_gtrack_create_sparse(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        intervals = pd.DataFrame(
            {
                "chrom": ["chr1", "chr1", "chr2"],
                "start": [0, 20, 10],
                "end": [10, 30, 20],
            }
        )
        pm.gtrack_create_sparse("created_sparse", "sparse test", intervals, [1.0, 2.0, 3.0])
        assert pm.gtrack_exists("created_sparse")
        info = pm.gtrack_info("created_sparse")
        assert info["type"] == "sparse"

        q = pd.DataFrame(
            {
                "chrom": ["chr1", "chr1", "chr2"],
                "start": [0, 20, 10],
                "end": [10, 30, 20],
            }
        )
        out = pm.gextract("created_sparse", q)
        assert out is not None
        np.testing.assert_allclose(out["created_sparse"].to_numpy(), np.array([1.0, 2.0, 3.0]), equal_nan=True)
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_create_dense_overlap_and_defval(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        intervals = pd.DataFrame(
            {
                "chrom": ["chr1", "chr1"],
                "start": [0, 5],
                "end": [10, 15],
            }
        )
        pm.gtrack_create_dense("created_dense", "dense test", intervals, [2.0, 4.0], binsize=10, defval=1.0)
        out = pm.gextract(
            "created_dense",
            pd.DataFrame({"chrom": ["chr1", "chr1"], "start": [0, 10], "end": [10, 20]}),
            iterator=10,
        )
        assert out is not None
        got = out["created_dense"].to_numpy(dtype=float, copy=False)
        np.testing.assert_allclose(got, np.array([40.0 / 15.0, 2.5]), rtol=1e-6, atol=1e-6)
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_import_bed_sparse_with_attrs(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        bed_path = tmp_path / "input.bed"
        bed_path.write_text(
            "chr1\t0\t10\ta\t5\n"
            "chr1\t10\t20\tb\t7\n",
            encoding="utf-8",
        )
        pm.gtrack_import(
            "imported_sparse",
            "import test",
            str(bed_path),
            binsize=0,
            attrs={"author": "tester"},
        )
        out = pm.gextract(
            "imported_sparse",
            pd.DataFrame({"chrom": ["chr1", "chr1"], "start": [0, 10], "end": [10, 20]}),
        )
        assert out is not None
        np.testing.assert_allclose(out["imported_sparse"].to_numpy(), np.array([5.0, 7.0]), equal_nan=True)
        assert pm.gtrack_attr_get("imported_sparse", "author") == "tester"
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_import_wig_fixedstep_dense(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        wig_path = tmp_path / "input.wig"
        wig_path.write_text(
            "track type=wiggle_0 name=test\n"
            "fixedStep chrom=chr1 start=1 step=10 span=10\n"
            "2\n"
            "4\n",
            encoding="utf-8",
        )
        pm.gtrack_import("imported_wig", "wig test", str(wig_path), binsize=10, defval=np.nan)
        out = pm.gextract(
            "imported_wig",
            pd.DataFrame({"chrom": ["chr1", "chr1"], "start": [0, 10], "end": [10, 20]}),
            iterator=10,
        )
        assert out is not None
        np.testing.assert_allclose(out["imported_wig"].to_numpy(dtype=float), np.array([2.0, 4.0]), equal_nan=True)
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_import_bigwig_optional(tmp_path):
    pybw = pytest.importorskip("pyBigWig")
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        bw_path = tmp_path / "input.bw"
        bw = pybw.open(str(bw_path), "w")
        try:
            bw.addHeader([("chr1", 1000)])
            bw.addEntries(["chr1", "chr1"], [0, 10], ends=[10, 20], values=[5.0, 7.0])
        finally:
            bw.close()

        pm.gtrack_import("imported_bw", "bigwig test", str(bw_path), binsize=0)
        out = pm.gextract(
            "imported_bw",
            pd.DataFrame({"chrom": ["chr1", "chr1"], "start": [0, 10], "end": [10, 20]}),
        )
        assert out is not None
        np.testing.assert_allclose(out["imported_bw"].to_numpy(dtype=float), np.array([5.0, 7.0]), equal_nan=True)
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_create_auto_indexed_db(tmp_path):
    root = _copy_db(tmp_path)
    try:
        seq_dir = root / "seq"
        (seq_dir / "genome.idx").write_bytes(b"")
        (seq_dir / "genome.seq").write_bytes(b"")
        pm.gdb_init(str(root))

        intervals = pd.DataFrame({"chrom": ["chr1"], "start": [0], "end": [10]})
        pm.gtrack_create_sparse("indexed_sparse", "indexed sparse", intervals, [1.0])

        tdir = root / "tracks" / "indexed_sparse.track"
        assert (tdir / "track.idx").exists()
        assert (tdir / "track.dat").exists()
        assert pm.gtrack_info("indexed_sparse")["format"] == "indexed"
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_create_expression_dense_streaming(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        pm.gtrack_create("created_expr_dense", "expr dense", "dense_track * 2", iterator=10)
        out = pm.gextract(
            "created_expr_dense",
            pd.DataFrame({"chrom": ["chr1"], "start": [0], "end": [100]}),
            iterator=10,
        )
        ref = pm.gextract(
            "dense_track * 2",
            pd.DataFrame({"chrom": ["chr1"], "start": [0], "end": [100]}),
            iterator=10,
        )
        assert out is not None and ref is not None
        np.testing.assert_allclose(
            out["created_expr_dense"].to_numpy(dtype=float),
            ref["dense_track * 2"].to_numpy(dtype=float),
            equal_nan=True,
        )
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_copy_mv_rm(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        pm.gtrack_copy("dense_track", "dense_copy")
        assert pm.gtrack_exists("dense_copy")

        pm.gtrack_mv("dense_copy", "dense_renamed")
        assert not pm.gtrack_exists("dense_copy")
        assert pm.gtrack_exists("dense_renamed")

        with pytest.raises(ValueError):
            pm.gtrack_rm("dense_renamed", force=False)
        pm.gtrack_rm("dense_renamed", force=True)
        assert not pm.gtrack_exists("dense_renamed")
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_import_mappedseq_sparse_dedup(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        inp = tmp_path / "mapped.tsv"
        inp.write_text(
            "A\tchr1\t10\t+\n"
            "A\tchr1\t10\t+\n"
            "A\tchr1\t20\t-\n",
            encoding="utf-8",
        )
        stats = pm.gtrack_import_mappedseq(
            "mapped_sparse",
            "mapped sparse",
            str(inp),
            pileup=0,
            binsize=-1,
            cols_order=(1, 2, 3, 4),
            remove_dups=True,
        )
        assert "total" in stats and "chrom" in stats
        out = pm.gextract(
            "mapped_sparse",
            pd.DataFrame({"chrom": ["chr1", "chr1"], "start": [10, 21], "end": [11, 22]}),
        )
        assert out is not None
        np.testing.assert_allclose(out["mapped_sparse"].to_numpy(dtype=float), np.array([1.0, 1.0]), equal_nan=True)
    finally:
        pm.gdb_init(str(TEST_DB))


def test_gtrack_import_mappedseq_dense(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        inp = tmp_path / "mapped_dense.tsv"
        inp.write_text(
            "AAA\tchr1\t10\t+\n"
            "AAA\tchr1\t30\t+\n",
            encoding="utf-8",
        )
        pm.gtrack_import_mappedseq(
            "mapped_dense",
            "mapped dense",
            str(inp),
            pileup=5,
            binsize=5,
            cols_order=(1, 2, 3, 4),
            remove_dups=True,
        )
        info = pm.gtrack_info("mapped_dense")
        assert info["type"] == "dense"
        out = pm.gextract(
            "mapped_dense",
            pd.DataFrame({"chrom": ["chr1", "chr1"], "start": [10, 30], "end": [15, 35]}),
            iterator=5,
        )
        assert out is not None
        vals = out["mapped_dense"].to_numpy(dtype=float)
        assert vals[0] > 0
        assert vals[1] > 0
    finally:
        pm.gdb_init(str(TEST_DB))


class TestGtrackImportRParity:
    """R misha parity tests ported from test-gtrack.import2/4/5.R."""

    def test_import_bed_dense_with_binsize(self, tmp_path):
        """Port of: import BED as dense track with binsize."""
        root = _copy_db(tmp_path)
        try:
            pm.gdb_init(str(root))
            bed = tmp_path / "dense_bed.bed"
            bed.write_text("chr1\t0\t10\tseg1\t2\nchr1\t10\t20\tseg2\t4\n")
            pm.gtrack_import("bed_dense", "BED dense track", str(bed), binsize=5, defval=0)
            info = pm.gtrack_info("bed_dense")
            assert info["type"] == "dense"
            assert int(info["bin_size"]) == 5
        finally:
            pm.gdb_init(str(TEST_DB))

    def test_import_tsv_sparse(self, tmp_path):
        """Port of: import tab-delimited with header chrom/start/end/value (sparse)."""
        root = _copy_db(tmp_path)
        try:
            pm.gdb_init(str(root))
            tsv = tmp_path / "sparse.tsv"
            tsv.write_text("chrom\tstart\tend\tvalue\nchr1\t0\t3\t2.5\nchr1\t4\t6\t1.0\n")
            pm.gtrack_import("tsv_sparse", "TSV import track", str(tsv), binsize=0)
            assert pm.gtrack_exists("tsv_sparse")
        finally:
            pm.gdb_init(str(TEST_DB))

    def test_import_tsv_multiple_value_cols_error(self, tmp_path):
        """Port of: import tab-delimited with multiple value columns fails."""
        root = _copy_db(tmp_path)
        try:
            pm.gdb_init(str(root))
            tsv_bad = tmp_path / "bad.tsv"
            tsv_bad.write_text("chrom\tstart\tend\tv1\tv2\nchr1\t0\t3\t2.5\t1.1\n")
            with pytest.raises(ValueError, match="value column"):
                pm.gtrack_import("tsv_bad", "bad TSV", str(tsv_bad), binsize=0)
        finally:
            pm.gdb_init(str(TEST_DB))

    def test_import_null_attrs(self, tmp_path):
        """Port of: import with attrs parameter - NULL attrs works."""
        root = _copy_db(tmp_path)
        try:
            pm.gdb_init(str(root))
            bed = tmp_path / "null_attr.bed"
            bed.write_text("chr1\t0\t10\ta\t5\n")
            pm.gtrack_import("null_attrs", "Test track", str(bed), binsize=0, attrs=None)
            assert pm.gtrack_exists("null_attrs")
        finally:
            pm.gdb_init(str(TEST_DB))

    def test_import_custom_attrs_with_defaults(self, tmp_path):
        """Port of: attrs parameter - attributes don't interfere with default attributes."""
        root = _copy_db(tmp_path)
        try:
            pm.gdb_init(str(root))
            bed = tmp_path / "custom_attr.bed"
            bed.write_text("chr1\t0\t10\ta\t5\n")
            pm.gtrack_import("custom_attrs", "Test description", str(bed), binsize=0,
                             attrs={"author": "test_user", "custom_attr": "custom_value"})
            assert pm.gtrack_attr_get("custom_attrs", "author") == "test_user"
            assert pm.gtrack_attr_get("custom_attrs", "custom_attr") == "custom_value"
            assert pm.gtrack_attr_get("custom_attrs", "description") == "Test description"
            assert len(pm.gtrack_attr_get("custom_attrs", "created.by")) > 0
            assert len(pm.gtrack_attr_get("custom_attrs", "created.date")) > 0
        finally:
            pm.gdb_init(str(TEST_DB))

    def test_import_multiple_attrs(self, tmp_path):
        """Port of: import with attrs parameter - multiple attributes."""
        root = _copy_db(tmp_path)
        try:
            pm.gdb_init(str(root))
            bed = tmp_path / "multi_attr.bed"
            bed.write_text("chr1\t0\t10\ta\t5\n")
            pm.gtrack_import("multi_attrs", "Test track", str(bed), binsize=0,
                             attrs={"author": "test_user", "version": "1.0", "experiment": "test_exp"})
            assert pm.gtrack_attr_get("multi_attrs", "author") == "test_user"
            assert pm.gtrack_attr_get("multi_attrs", "version") == "1.0"
            assert pm.gtrack_attr_get("multi_attrs", "experiment") == "test_exp"
        finally:
            pm.gdb_init(str(TEST_DB))


def test_gtrack_import_set_mixed_success(tmp_path):
    root = _copy_db(tmp_path)
    try:
        pm.gdb_init(str(root))
        good = tmp_path / "good.bed"
        bad = tmp_path / "bad.bed"
        good.write_text("chr1\t0\t10\ta\t3\n", encoding="utf-8")
        bad.write_text("chr1\tbad\t10\ta\t3\n", encoding="utf-8")

        res = pm.gtrack_import_set(
            description="batch import",
            path=str(tmp_path / "*.bed"),
            binsize=0,
            track_prefix="batch_",
            defval=np.nan,
        )
        assert "good.bed" in res.get("files_imported", [])
        assert "bad.bed" in res.get("files_failed", [])
        assert pm.gtrack_exists("batch_good")
        assert not pm.gtrack_exists("batch_bad")
    finally:
        pm.gdb_init(str(TEST_DB))



"""Tests for gseq_kmer and gseq_kmer_dist."""

import numpy as np
import pandas as pd
import pytest

import pymisha as pm


@pytest.fixture(autouse=True)
def _ensure_db(_init_db):
    pass


class TestGseqKmer:
    """Tests for gseq_kmer function."""

    def test_basic_count(self):
        """Count a simple k-mer on forward strand."""
        result = pm.gseq_kmer(["ACGTACGT"], "ACG", mode="count", strand=1)
        assert result[0] == 2  # ACG appears at positions 0 and 4

    def test_single_base_count(self):
        """Count single base occurrences."""
        result = pm.gseq_kmer(["AAACCC"], "A", mode="count", strand=1)
        assert result[0] == 3

    def test_forward_strand_only(self):
        """Count on forward strand only."""
        result = pm.gseq_kmer(["ACGTACGT"], "ACG", mode="count", strand=1)
        assert result[0] == 2

    def test_reverse_strand_only(self):
        """Count on reverse strand only."""
        # ACG reverse complement is CGT
        result = pm.gseq_kmer(["ACGTACGT"], "ACG", mode="count", strand=-1)
        # CGT appears at positions 2 and 6
        assert result[0] == 2

    def test_both_strands(self):
        """Count on both strands."""
        # R misha counts both strands even for palindromic kmers.
        result = pm.gseq_kmer(["AACGTT"], "CG", mode="count", strand=0)
        assert result[0] == 2

    def test_non_palindrome_both_strands(self):
        """Non-palindromic k-mer counted on both strands."""
        # ACG fwd at pos 0, CGT (revcomp ACG) at pos 1
        result = pm.gseq_kmer(["ACGT"], "ACG", mode="count", strand=0)
        assert result[0] == 2  # ACG forward + CGT (=revcomp ACG) at pos 1

    def test_fraction_mode(self):
        """Fraction mode returns count / possible positions."""
        # "ACGTACGT" has 8 chars, k=3, so 6 possible positions
        result = pm.gseq_kmer(["ACGTACGT"], "ACG", mode="frac", strand=1)
        assert abs(result[0] - 2.0 / 6) < 1e-10

    def test_multiple_sequences(self):
        """Works with multiple input sequences."""
        result = pm.gseq_kmer(["AACG", "CCCC", "ACGA"], "CG", mode="count", strand=1)
        assert len(result) == 3
        assert result[0] == 1  # CG at pos 2
        assert result[1] == 0  # no CG
        assert result[2] == 1  # CG at pos 1

    def test_string_input(self):
        """Single string input works."""
        result = pm.gseq_kmer("ACGT", "ACG", mode="count", strand=1)
        assert len(result) == 1
        assert result[0] == 1

    def test_empty_sequence(self):
        """Empty sequence returns 0."""
        result = pm.gseq_kmer([""], "ACG", mode="count")
        assert result[0] == 0

    def test_kmer_longer_than_seq(self):
        """K-mer longer than sequence returns 0."""
        result = pm.gseq_kmer(["AC"], "ACGT", mode="count")
        assert result[0] == 0

    def test_start_end_pos(self):
        """ROI boundaries limit search."""
        # "ACGTACGT" (1-indexed: A=1, C=2, G=3, T=4, A=5, C=6, G=7, T=8)
        # ROI 5-8 = "ACGT", ACG at pos 0 → 1 match
        result = pm.gseq_kmer(["ACGTACGT"], "ACG", mode="count", strand=1,
                               start_pos=5, end_pos=8)
        assert result[0] == 1
        # Full sequence has 2 ACG matches
        result_full = pm.gseq_kmer(["ACGTACGT"], "ACG", mode="count", strand=1)
        assert result_full[0] == 2

    def test_invalid_kmer(self):
        """Invalid k-mer characters raise error."""
        with pytest.raises(ValueError, match="only A, C, G, T"):
            pm.gseq_kmer(["ACGT"], "AXG")

    def test_invalid_strand(self):
        """Invalid strand value raises error."""
        with pytest.raises(ValueError, match="strand must be"):
            pm.gseq_kmer(["ACGT"], "ACG", strand=2)

    def test_case_insensitive(self):
        """Case-insensitive matching."""
        result = pm.gseq_kmer(["acgtACGT"], "ACG", mode="count", strand=1)
        assert result[0] == 2

    def test_gap_skipping(self):
        """Gap characters are skipped."""
        # "AC-GT" with gaps removed becomes "ACGT"
        result = pm.gseq_kmer(["AC-GT"], "ACGT", mode="count", strand=1,
                               skip_gaps=True, gap_chars=["-"])
        assert result[0] == 1

    def test_returns_numpy_array(self):
        """Return type is numpy array."""
        result = pm.gseq_kmer(["ACGT"], "AC")
        assert isinstance(result, np.ndarray)


class TestGseqKmerDist:
    """Tests for gseq_kmer_dist function."""

    def test_basic_kmer_dist(self):
        """Count k-mer distribution in a genomic region."""
        intervals = pm.gintervals("1", 0, 1000)
        result = pm.gseq_kmer_dist(intervals, k=2)
        assert isinstance(result, pd.DataFrame)
        assert "kmer" in result.columns
        assert "count" in result.columns
        assert len(result) > 0
        # All k-mers should be of length 2
        assert all(len(k) == 2 for k in result["kmer"])
        # Counts should be positive
        assert (result["count"] > 0).all()

    def test_kmer_dist_sorted(self):
        """K-mers should be sorted alphabetically."""
        intervals = pm.gintervals("1", 0, 1000)
        result = pm.gseq_kmer_dist(intervals, k=2)
        kmers = result["kmer"].tolist()
        assert kmers == sorted(kmers)

    def test_kmer_dist_k1(self):
        """k=1 should return at most 4 k-mers (A, C, G, T)."""
        intervals = pm.gintervals("1", 0, 1000)
        result = pm.gseq_kmer_dist(intervals, k=1)
        assert len(result) <= 4
        assert set(result["kmer"]).issubset({"A", "C", "G", "T"})

    def test_kmer_dist_large_k(self):
        """k=6 returns many k-mers."""
        intervals = pm.gintervals("1", 0, 10000)
        result = pm.gseq_kmer_dist(intervals, k=6)
        assert len(result) > 10  # Should have many 6-mers in 10kb

    def test_kmer_dist_with_mask(self):
        """Mask intervals exclude regions from counting."""
        intervals = pm.gintervals("1", 0, 1000)
        mask = pm.gintervals("1", 0, 500)

        full_result = pm.gseq_kmer_dist(intervals, k=2)
        masked_result = pm.gseq_kmer_dist(intervals, k=2, mask=mask)

        # Masked result should have fewer total counts
        full_total = full_result["count"].sum()
        masked_total = masked_result["count"].sum()
        assert masked_total < full_total

    def test_kmer_dist_invalid_k(self):
        """Invalid k raises ValueError."""
        intervals = pm.gintervals("1", 0, 1000)
        with pytest.raises(ValueError, match="k must be"):
            pm.gseq_kmer_dist(intervals, k=0)
        with pytest.raises(ValueError, match="k must be"):
            pm.gseq_kmer_dist(intervals, k=11)

    def test_kmer_dist_total_matches_sequence(self):
        """Total k-mer count should match expected from sequence length."""
        intervals = pm.gintervals("1", 0, 100)
        result = pm.gseq_kmer_dist(intervals, k=1)
        total = result["count"].sum()
        # For k=1, total should equal sequence length (minus any N bases)
        seqs = pm.gseq_extract(intervals)
        expected = sum(1 for c in seqs[0].upper() if c in "ACGT")
        assert total == expected



import shutil
import subprocess

import pandas as pd
import pytest

import pymisha as pm


def _require_rscript():
    if shutil.which("Rscript") is None:
        pytest.skip("Rscript not available")


def _write_r_serialized_df(path, r_df_expr):
    r_cmd = (
        "out<-commandArgs(TRUE)[1]; "
        f"df<-{r_df_expr}; "
        "con<-file(out,'wb'); serialize(df, con); close(con)"
    )
    subprocess.check_call(["Rscript", "-e", r_cmd, str(path)])


def _write_meta(path, stats_expr, zeroline_expr):
    r_cmd = (
        "out<-commandArgs(TRUE)[1]; "
        f"stats<-{stats_expr}; "
        f"zeroline<-{zeroline_expr}; "
        "con<-file(out,'wb'); serialize(list(stats=stats, zeroline=zeroline), con); close(con)"
    )
    subprocess.check_call(["Rscript", "-e", r_cmd, str(path)])


def _create_db_root(tmp_path):
    root = tmp_path / "db"
    root.mkdir()
    chrom_sizes = root / "chrom_sizes.txt"
    chrom_sizes.write_text("1\t100\n2\t100\n")
    tracks = root / "tracks"
    tracks.mkdir()
    return root


@pytest.fixture
def restore_db():
    current_root = pm._shared._GROOT
    yield
    if current_root:
        pm.gdb_init(current_root)

def test_indexed_bigset_1d_roundtrip(tmp_path, restore_db):
    _require_rscript()
    root = _create_db_root(tmp_path)
    interv_dir = root / "tracks" / "big1.interv"
    interv_dir.mkdir()

    _write_r_serialized_df(
        interv_dir / "chr1",
        "data.frame(chrom=factor(c('chr1','chr1')), start=c(0,10), end=c(5,20))",
    )
    _write_r_serialized_df(
        interv_dir / "chr2",
        "data.frame(chrom=factor(c('chr2')), start=c(3), end=c(8))",
    )
    _write_meta(
        interv_dir / ".meta",
        "data.frame(chrom=factor(c('chr1','chr2')), size=c(2,1))",
        "data.frame(chrom=factor(character()), start=integer(), end=integer())",
    )

    pm.gdb_init(str(root))
    df = pm.gintervals_load("big1")
    assert df is not None
    assert len(df) == 3

    pm.gintervals_convert_to_indexed("big1", remove_old=True, force=True)
    assert pm.gintervals_is_indexed("big1")

    df2 = pm.gintervals_load("big1")
    assert len(df2) == 3
    pd.testing.assert_frame_equal(
        df.sort_values(["chrom", "start"]).reset_index(drop=True),
        df2.sort_values(["chrom", "start"]).reset_index(drop=True),
        check_dtype=False,
    )

    df_chr1 = pm.gintervals_load("big1", chrom="chr1")
    assert len(df_chr1) == 2


def test_indexed_bigset_2d_roundtrip(tmp_path, restore_db):
    _require_rscript()
    root = _create_db_root(tmp_path)
    interv_dir = root / "tracks" / "big2d.interv"
    interv_dir.mkdir()

    _write_r_serialized_df(
        interv_dir / "chr1-chr1",
        "data.frame(chrom1=factor(c('chr1')), start1=c(0), end1=c(10), "
        "chrom2=factor(c('chr1')), start2=c(5), end2=c(15))",
    )
    _write_meta(
        interv_dir / ".meta",
        "data.frame(chrom1=factor(c('chr1')), chrom2=factor(c('chr1')), size=c(1))",
        "data.frame(chrom1=factor(character()), start1=integer(), end1=integer(), "
        "chrom2=factor(character()), start2=integer(), end2=integer())",
    )

    pm.gdb_init(str(root))
    df = pm.gintervals_load("big2d")
    assert df is not None
    assert len(df) == 1

    pm.gintervals_2d_convert_to_indexed("big2d", remove_old=True, force=True)
    assert pm.gintervals_is_indexed("big2d")

    df2 = pm.gintervals_load("big2d")
    assert len(df2) == 1
    pd.testing.assert_frame_equal(
        df.sort_values(["chrom1", "start1"]).reset_index(drop=True),
        df2.sort_values(["chrom1", "start1"]).reset_index(drop=True),
        check_dtype=False,
    )

    df_pair = pm.gintervals_load("big2d", chrom1="chr1", chrom2="chr1")
    assert len(df_pair) == 1



"""Tests for gvtrack_iterator_2d."""

import pytest

import pymisha as pm


@pytest.fixture(autouse=True)
def _ensure_db(_init_db):
    pass


@pytest.fixture(autouse=True)
def _clean_vtracks():
    yield
    pm.gvtrack_clear()


class TestGvtrackIterator2d:
    """Tests for gvtrack_iterator_2d."""

    def test_sets_2d_iterator_defaults(self):
        """Setting 2D iterator with defaults stores zero shifts."""
        pm.gvtrack_create("vt1", "dense_track")
        pm.gvtrack_iterator_2d("vt1")
        info = pm.gvtrack_info("vt1")
        assert info.get("itr_type") == "2d"
        assert info.get("sshift1") == 0
        assert info.get("eshift1") == 0
        assert info.get("sshift2") == 0
        assert info.get("eshift2") == 0

    def test_sets_custom_shifts(self):
        """Setting 2D iterator with custom shifts stores them."""
        pm.gvtrack_create("vt2", "dense_track")
        pm.gvtrack_iterator_2d("vt2", sshift1=-100, eshift1=100, sshift2=-200, eshift2=200)
        info = pm.gvtrack_info("vt2")
        assert info["itr_type"] == "2d"
        assert info["sshift1"] == -100
        assert info["eshift1"] == 100
        assert info["sshift2"] == -200
        assert info["eshift2"] == 200

    def test_nonexistent_vtrack_raises(self):
        """Setting iterator on nonexistent vtrack raises KeyError."""
        with pytest.raises(KeyError):
            pm.gvtrack_iterator_2d("nonexistent")

    def test_overwrites_previous_iterator(self):
        """Setting 2D iterator replaces any previous 1D iterator."""
        pm.gvtrack_create("vt3", "dense_track")
        pm.gvtrack_iterator("vt3", sshift=-50, eshift=50)
        info1 = pm.gvtrack_info("vt3")
        assert info1.get("sshift") == -50

        pm.gvtrack_iterator_2d("vt3", sshift1=10, eshift2=20)
        info2 = pm.gvtrack_info("vt3")
        assert info2["itr_type"] == "2d"
        assert info2["sshift1"] == 10
        assert info2["eshift2"] == 20



"""Tests for gdb_info (database inspection metadata)."""

import pandas as pd
import pytest

import pymisha as pm


def test_gdb_info_uses_active_db():
    """Without a path, gdb_info should inspect the active DB."""
    info = pm.gdb_info()
    assert info["is_db"] is True
    assert info["format"] in {"indexed", "per-chromosome"}
    assert info["num_chromosomes"] > 0
    assert info["genome_size"] > 0
    assert isinstance(info["chromosomes"], pd.DataFrame)
    assert list(info["chromosomes"].columns) == ["chrom", "size"]


def test_gdb_info_nonexistent_dir():
    """Nonexistent paths should return is_db=False with an error message."""
    info = pm.gdb_info("/tmp/does_not_exist_pymisha_12345")
    assert info["is_db"] is False
    assert "does not exist" in info["error"]


def test_gdb_info_not_misha_db(tmp_path):
    """Directories without chrom_sizes.txt are not considered misha DBs."""
    root = tmp_path / "notadb"
    root.mkdir()

    info = pm.gdb_info(str(root))
    assert info["is_db"] is False
    assert "chrom_sizes.txt" in info["error"]


def test_gdb_info_per_chromosome_format(tmp_path):
    """DB with chrom_sizes but without seq/genome.idx+seq is per-chromosome."""
    root = tmp_path / "db"
    (root / "seq").mkdir(parents=True)
    (root / "tracks").mkdir(parents=True)
    (root / "chrom_sizes.txt").write_text("chr1\t1000\nchr2\t2000\n", encoding="utf-8")

    info = pm.gdb_info(str(root))
    assert info["is_db"] is True
    assert info["format"] == "per-chromosome"
    assert info["num_chromosomes"] == 2
    assert info["genome_size"] == 3000


def test_gdb_info_indexed_format(tmp_path):
    """DB with seq/genome.idx and seq/genome.seq is indexed."""
    root = tmp_path / "db"
    seq_dir = root / "seq"
    seq_dir.mkdir(parents=True)
    (root / "tracks").mkdir(parents=True)
    (root / "chrom_sizes.txt").write_text("chr1\t1000\n", encoding="utf-8")
    (seq_dir / "genome.idx").write_bytes(b"idx")
    (seq_dir / "genome.seq").write_bytes(b"seq")

    info = pm.gdb_info(str(root))
    assert info["is_db"] is True
    assert info["format"] == "indexed"


def test_gdb_info_requires_active_db_when_path_not_given():
    """If no active DB and no path, function should raise."""
    old_root = pm._shared._GROOT
    try:
        pm._shared._GROOT = None
        with pytest.raises(ValueError, match="No database is currently active"):
            pm.gdb_info()
    finally:
        pm._shared._GROOT = old_root


# ============================================================================
# Tests for gtrack_info (ported from R test-gtrack.info.R)
# ============================================================================

class TestGtrackInfo:
    """Tests for gtrack_info function.

    Ported from R misha test-gtrack.info.R.
    Test DB tracks: dense_track, sparse_track, array_track, rects_track.
    """

    def test_gtrack_info_dense_track(self):
        """gtrack_info returns correct metadata for a dense track."""
        info = pm.gtrack_info("dense_track")
        assert info["type"] == "dense"
        assert info["dimensions"] == 1
        assert info["format"] == "per-chromosome"
        assert info["size_in_bytes"] > 0

    def test_gtrack_info_dense_track_bin_size(self):
        """Dense track has bin_size in gtrack_info result."""
        info = pm.gtrack_info("dense_track")
        assert "bin_size" in info
        assert info["bin_size"] == 50

    def test_gtrack_info_sparse_track(self):
        """gtrack_info returns correct metadata for a sparse track."""
        info = pm.gtrack_info("sparse_track")
        assert info["type"] == "sparse"
        assert info["dimensions"] == 1
        assert info["format"] == "per-chromosome"
        assert info["size_in_bytes"] > 0

    def test_gtrack_info_sparse_track_no_bin_size(self):
        """Sparse track does not have bin_size in gtrack_info result."""
        info = pm.gtrack_info("sparse_track")
        assert "bin_size" not in info

    def test_gtrack_info_array_track(self):
        """gtrack_info returns correct metadata for an array track."""
        info = pm.gtrack_info("array_track")
        assert info["type"] == "array"
        assert info["dimensions"] == 1
        assert info["format"] == "per-chromosome"
        assert info["size_in_bytes"] > 0

    def test_gtrack_info_2d_rects_track(self):
        """gtrack_info returns correct metadata for a 2D rectangles track."""
        info = pm.gtrack_info("rects_track")
        assert info["type"] == "rectangles"
        assert info["dimensions"] == 2
        assert info["format"] == "per-chromosome"

    def test_gtrack_info_dense_track2_in_subdir(self):
        """gtrack_info works for tracks in subdirectories."""
        info = pm.gtrack_info("subdir.dense_track2")
        assert info["type"] == "dense"
        assert info["dimensions"] == 1
        assert "bin_size" in info
        assert info["bin_size"] == 50

    def test_gtrack_info_returns_attributes(self):
        """gtrack_info includes track attributes."""
        info = pm.gtrack_info("dense_track")
        assert "attributes" in info
        assert isinstance(info["attributes"], dict)
        # Dense track should have created.by and created.date
        assert "created.by" in info["attributes"]
        assert "created.date" in info["attributes"]

    def test_gtrack_info_nonexistent_track_raises(self):
        """gtrack_info raises error for non-existent track."""
        with pytest.raises(Exception):
            pm.gtrack_info("nonexistent_track_xyz")



"""Tests for 2D track creation, import, and reading."""

import os
import shutil
import struct

import numpy as np
import pytest

import pymisha as pm

TRACK_DIR = os.path.join(
    os.path.dirname(__file__), "testdb", "trackdb", "test", "tracks"
)


def _track_dir(name):
    """Get track directory path (dots become subdirectories)."""
    return os.path.join(TRACK_DIR, name.replace(".", "/") + ".track")


def _cleanup_track(name):
    """Remove a track and reload DB."""
    tdir = _track_dir(name)
    if os.path.exists(tdir):
        shutil.rmtree(tdir)
        import _pymisha
        _pymisha.pm_dbreload()


def test_quadtree_split_preserves_parent_stats():
    from pymisha._quadtree import QuadTree

    qtree = QuadTree(0, 0, 16, 16, max_node_objs=1)
    qtree.insert((0, 0, 8, 8, 10.0))
    qtree.insert((8, 8, 16, 16, 20.0))

    root = qtree.root
    assert not root.is_leaf
    assert root.stat["occupied_area"] == 128
    assert root.stat["weighted_sum"] == pytest.approx(1920.0)
    assert root.stat["min_val"] == 10.0
    assert root.stat["max_val"] == 20.0


class TestGtrack2dCreate:
    """Tests for gtrack_2d_create."""

    def _cleanup_track(self, name):
        _cleanup_track(name)

    def test_create_basic_rects(self):
        """Create a simple 2D rectangles track."""
        tname = "test.test_2d_rects"
        self._cleanup_track(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1", "1"],
                starts1=[100, 200000],
                ends1=[200, 300000],
                chroms2=["1", "1"],
                starts2=[300, 400000],
                ends2=[400, 499000],
            )
            values = [1.5, 2.5]
            pm.gtrack_2d_create(tname, "test 2D track", intervals, values)

            assert pm.gtrack_exists(tname)
            info = pm.gtrack_info(tname)
            assert info["type"] == "rectangles"
        finally:
            self._cleanup_track(tname)

    def test_create_points_track(self):
        """Create a 2D points track (all intervals are 1bp)."""
        tname = "test.test_2d_points"
        self._cleanup_track(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1", "1"],
                starts1=[100, 200],
                ends1=[101, 201],
                chroms2=["1", "1"],
                starts2=[300, 400],
                ends2=[301, 401],
            )
            values = [10.0, 20.0]
            pm.gtrack_2d_create(tname, "test points", intervals, values)

            assert pm.gtrack_exists(tname)
            info = pm.gtrack_info(tname)
            assert info["type"] == "points"
        finally:
            self._cleanup_track(tname)

    def test_create_multi_chrom_pairs(self):
        """Create 2D track spanning multiple chromosome pairs."""
        tname = "test.test_2d_multi"
        self._cleanup_track(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1", "1", "2"],
                starts1=[100, 100, 50],
                ends1=[200, 200, 150],
                chroms2=["1", "2", "2"],
                starts2=[300, 100, 200],
                ends2=[400, 200, 299000],
            )
            values = [1.0, 2.0, 3.0]
            pm.gtrack_2d_create(tname, "multi chrom", intervals, values)

            assert pm.gtrack_exists(tname)
            # Check that per-chrom-pair files exist
            tdir = _track_dir(tname)
            # Files should be named with normalized chroms
            files = [f for f in os.listdir(tdir) if not f.startswith(".")]
            assert len(files) >= 2  # at least 1-1 and 1-2 (or 2-2)
        finally:
            self._cleanup_track(tname)

    def test_create_values_length_mismatch(self):
        """Values length must match intervals length."""
        intervals = pm.gintervals_2d(
            chroms1=["1"], starts1=[100], ends1=[200],
            chroms2=["1"], starts2=[300], ends2=[400],
        )
        with pytest.raises(ValueError, match="values"):
            pm.gtrack_2d_create("test.bad", "bad", intervals, [1.0, 2.0])

    def test_create_overlapping_intervals_error(self):
        """Overlapping 2D intervals should raise an error."""
        intervals = pm.gintervals_2d(
            chroms1=["1", "1"],
            starts1=[100, 150],
            ends1=[200, 250],
            chroms2=["1", "1"],
            starts2=[300, 350],
            ends2=[400, 450],
        )
        tname = "test.test_2d_overlap"
        self._cleanup_track(tname)
        try:
            with pytest.raises(ValueError, match="[Oo]verlap"):
                pm.gtrack_2d_create(tname, "overlap", intervals, [1.0, 2.0])
        finally:
            self._cleanup_track(tname)

    def test_roundtrip_read_existing_track(self):
        """Read the existing rects_track from test DB and verify we can parse it."""
        # The test DB has rects_track with chr1-chr1, chr1-chr2, chr1-chrX
        info = pm.gtrack_info("rects_track")
        assert info["type"] == "rectangles"


class TestGtrack2dImport:
    """Tests for gtrack_2d_import."""

    def _cleanup_track(self, name):
        _cleanup_track(name)

    def test_import_from_file(self, tmp_path):
        """Import 2D track from tab-delimited file."""
        tname = "test.test_2d_import"
        self._cleanup_track(tname)
        try:
            # Write a tab-delimited file
            tsv = tmp_path / "contacts.tsv"
            tsv.write_text(
                "chrom1\tstart1\tend1\tchrom2\tstart2\tend2\tvalue\n"
                "1\t100\t200\t1\t300\t400\t5.5\n"
                "1\t1000\t2000\t2\t5000\t6000\t3.0\n"
            )
            pm.gtrack_2d_import(tname, "imported 2D", str(tsv))
            assert pm.gtrack_exists(tname)
            info = pm.gtrack_info(tname)
            assert info["type"] == "rectangles"
        finally:
            self._cleanup_track(tname)

    def test_import_points_from_file(self, tmp_path):
        """Import 2D points track from tab-delimited file."""
        tname = "test.test_2d_import_pts"
        self._cleanup_track(tname)
        try:
            tsv = tmp_path / "points.tsv"
            tsv.write_text(
                "chrom1\tstart1\tend1\tchrom2\tstart2\tend2\tvalue\n"
                "1\t100\t101\t1\t300\t301\t5.5\n"
                "1\t1000\t1001\t2\t5000\t5001\t3.0\n"
            )
            pm.gtrack_2d_import(tname, "imported points", str(tsv))
            assert pm.gtrack_exists(tname)
            info = pm.gtrack_info(tname)
            assert info["type"] == "points"
        finally:
            self._cleanup_track(tname)


class TestQuadTreeBinaryCompat:
    """Test that our quad-tree writer produces files readable by the existing rects_track reader."""

    def test_create_and_verify_binary_format(self):
        """Create a track and verify the binary file has correct signature."""
        tname = "test.test_2d_bincompat"
        _cleanup_track(tname)
        tdir = _track_dir(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1"],
                starts1=[100],
                ends1=[200],
                chroms2=["1"],
                starts2=[300],
                ends2=[400],
            )
            pm.gtrack_2d_create(tname, "bincompat", intervals, [42.0])

            # Check binary format
            chrom_file = os.path.join(tdir, "1-1")
            assert os.path.exists(chrom_file), f"Expected 1-1 in {os.listdir(tdir)}"
            with open(chrom_file, "rb") as f:
                data = f.read()

            # Signature should be -9 (RECTS)
            sig = struct.unpack_from("<i", data, 0)[0]
            assert sig == -9

            # num_objs
            num_objs = struct.unpack_from("<Q", data, 4)[0]
            assert num_objs == 1

            # root_chunk_fpos
            root_fpos = struct.unpack_from("<q", data, 12)[0]
            assert root_fpos > 0

            # Read chunk header
            chunk_size, top_node_off = struct.unpack_from("<qq", data, root_fpos)
            assert chunk_size > 0
            assert top_node_off > 0

            # Read leaf
            node_start = root_fpos + top_node_off
            is_leaf = data[node_start]
            assert is_leaf == 1

            # Read stat
            off = node_start + 8  # skip is_leaf + padding
            occ_area = struct.unpack_from("<q", data, off)[0]
            assert occ_area == 100 * 100  # (200-100) * (400-300)

            # Read arena
            off += 32  # skip stat
            x1, y1, x2, y2 = struct.unpack_from("<qqqq", data, off)
            assert x1 == 0 and y1 == 0  # arena is full chrom range

            # Read num_objs in leaf
            off += 32
            n = struct.unpack_from("<I", data, off)[0]
            assert n == 1

            # Read object
            off += 8  # 4 bytes + 4 padding
            obj_id = struct.unpack_from("<Q", data, off)[0]
            assert obj_id == 0
            rx1, ry1, rx2, ry2 = struct.unpack_from("<qqqq", data, off + 8)
            assert (rx1, ry1, rx2, ry2) == (100, 300, 200, 400)
            rv = struct.unpack_from("<f", data, off + 40)[0]
            assert abs(rv - 42.0) < 0.01
        finally:
            _cleanup_track(tname)

    def test_points_binary_format(self):
        """Create a points track and verify signature is -10."""
        tname = "test.test_2d_pts_bin"
        _cleanup_track(tname)
        tdir = _track_dir(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1"],
                starts1=[100],
                ends1=[101],
                chroms2=["1"],
                starts2=[300],
                ends2=[301],
            )
            pm.gtrack_2d_create(tname, "pts bin", intervals, [7.0])

            chrom_file = os.path.join(tdir, "1-1")
            with open(chrom_file, "rb") as f:
                sig = struct.unpack_from("<i", f.read(4), 0)[0]
            assert sig == -10  # POINTS format
        finally:
            _cleanup_track(tname)

    def test_cpp_reads_created_rects_track(self):
        """Verify C++ type detection works on Python-created RECTS track."""
        tname = "test.test_2d_cpp_read"
        _cleanup_track(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1", "1", "2"],
                starts1=[100, 10000, 50],
                ends1=[200, 20000, 150],
                chroms2=["1", "1", "2"],
                starts2=[300, 30000, 200],
                ends2=[400, 40000, 250],
            )
            pm.gtrack_2d_create(tname, "cpp read test", intervals, [1.0, 2.0, 3.0])
            info = pm.gtrack_info(tname)
            assert info["type"] == "rectangles"
            assert info["dimensions"] == 2
        finally:
            _cleanup_track(tname)

    def test_cpp_reads_created_points_track(self):
        """Verify C++ type detection works on Python-created POINTS track."""
        tname = "test.test_2d_cpp_pts"
        _cleanup_track(tname)
        try:
            intervals = pm.gintervals_2d(
                chroms1=["1"],
                starts1=[100],
                ends1=[101],
                chroms2=["1"],
                starts2=[300],
                ends2=[301],
            )
            pm.gtrack_2d_create(tname, "cpp pts test", intervals, [5.0])
            info = pm.gtrack_info(tname)
            assert info["type"] == "points"
            assert info["dimensions"] == 2
        finally:
            _cleanup_track(tname)

    def test_many_objects_trigger_splitting(self):
        """Create a track with enough objects to trigger quad-tree splitting."""
        tname = "test.test_2d_split"
        _cleanup_track(tname)
        try:
            n = 100
            rng = np.random.RandomState(42)
            starts1 = rng.randint(0, 400000, size=n)
            starts2 = rng.randint(0, 400000, size=n)
            # Make small non-overlapping rectangles
            ends1 = starts1 + 10
            ends2 = starts2 + 10
            vals = rng.uniform(1, 100, size=n).astype(np.float32)

            intervals = pm.gintervals_2d(
                chroms1=["1"] * n,
                starts1=starts1.tolist(),
                ends1=ends1.tolist(),
                chroms2=["1"] * n,
                starts2=starts2.tolist(),
                ends2=ends2.tolist(),
            )
            pm.gtrack_2d_create(tname, "split test", intervals, vals.tolist())
            info = pm.gtrack_info(tname)
            assert info["type"] == "rectangles"
            assert info["dimensions"] == 2
        finally:
            _cleanup_track(tname)



"""Tests for glookup function."""

import numpy as np
import pandas as pd
import pytest

import pymisha as pm


class TestGlookupBasic:
    """Test basic glookup functionality."""

    def test_glookup_1d_lookup_table(self):
        """One-dimensional lookup table with track expression."""
        # Create a 1D lookup table with 5 values
        lookup_table = np.array([10, 20, 30, 40, 50])
        # Breaks create 5 bins: (0.1, 0.12], (0.12, 0.14], ..., (0.18, 0.2]
        breaks = np.linspace(0.1, 0.2, 6)

        # Get intervals - dense_track has values roughly 0.1-0.2
        intervals = pm.gintervals("1", 0, 200)

        result = pm.glookup(lookup_table, "dense_track", breaks, intervals=intervals)

        assert result is not None
        assert isinstance(result, pd.DataFrame)
        assert "chrom" in result.columns
        assert "start" in result.columns
        assert "end" in result.columns
        assert "value" in result.columns

        # All values should be from the lookup table
        assert all(v in lookup_table or np.isnan(v) for v in result["value"])

    def test_glookup_2d_lookup_table(self):
        """Two-dimensional lookup table with two expressions."""
        # Create a 2D lookup table (5 x 3)
        lookup_table = np.arange(1, 16).reshape((5, 3))
        breaks1 = np.linspace(0.1, 0.2, 6)  # 5 bins for first dim
        breaks2 = np.linspace(0.31, 0.37, 4)  # 3 bins for second dim

        intervals = pm.gintervals("1", 0, 200)

        result = pm.glookup(
            lookup_table,
            "dense_track", breaks1,
            "2 * dense_track", breaks2,
            intervals=intervals
        )

        assert result is not None
        assert isinstance(result, pd.DataFrame)
        assert "value" in result.columns

    def test_glookup_returns_intervalID(self):
        """Result has intervalID column for tracking."""
        lookup_table = np.array([10, 20, 30])
        breaks = np.array([0.0, 0.15, 0.3, 0.45])
        intervals = pm.gintervals("1", [0, 100], [100, 200])

        result = pm.glookup(lookup_table, "dense_track", breaks, intervals=intervals)

        assert result is not None
        assert "intervalID" in result.columns


class TestGlookupBinning:
    """Test binning behavior."""

    def test_glookup_include_lowest(self):
        """include_lowest includes the lowest break value in first bin."""
        # Values exactly at lowest break go to bin 0 when include_lowest=True
        lookup_table = np.array([100, 200, 300])
        breaks = np.array([0.1, 0.15, 0.2, 0.25])

        intervals = pm.gintervals("1", 0, 100)

        result_with = pm.glookup(
            lookup_table, "dense_track", breaks,
            intervals=intervals, include_lowest=True
        )
        result_without = pm.glookup(
            lookup_table, "dense_track", breaks,
            intervals=intervals, include_lowest=False
        )

        # Results should differ when values exactly equal lowest break
        assert result_with is not None
        assert result_without is not None

    def test_glookup_force_binning_true(self):
        """force_binning=True clamps out-of-range values to first/last bin."""
        lookup_table = np.array([1, 2, 3])
        # Narrow breaks that will miss many values
        breaks = np.array([0.14, 0.15, 0.16, 0.17])

        intervals = pm.gintervals("1", 0, 500)

        result = pm.glookup(
            lookup_table, "dense_track", breaks,
            intervals=intervals, force_binning=True
        )

        assert result is not None
        # With force_binning, no NaN values (out-of-range are clamped)
        # All values should be from lookup_table
        valid_values = result["value"].dropna()
        assert len(valid_values) > 0

    def test_glookup_force_binning_false(self):
        """force_binning=False produces NaN for out-of-range values."""
        lookup_table = np.array([1, 2, 3])
        # Very narrow breaks - most values will be out of range
        breaks = np.array([0.14, 0.145, 0.15, 0.155])

        intervals = pm.gintervals("1", 0, 500)

        result = pm.glookup(
            lookup_table, "dense_track", breaks,
            intervals=intervals, force_binning=False
        )

        assert result is not None
        # With force_binning=False, out-of-range values become NaN
        # There should be some NaN values
        assert result["value"].isna().any()


class TestGlookupEdgeCases:
    """Test edge cases."""

    def test_glookup_empty_intervals(self):
        """Empty intervals returns None or empty result."""
        lookup_table = np.array([1, 2, 3])
        breaks = np.array([0.0, 0.1, 0.2, 0.3])
        intervals = pd.DataFrame(columns=["chrom", "start", "end"])

        result = pm.glookup(lookup_table, "dense_track", breaks, intervals=intervals)

        assert result is None or len(result) == 0

    def test_glookup_nan_values_in_expression(self):
        """NaN values in expression produce NaN in result."""
        lookup_table = np.array([1, 2, 3])
        breaks = np.array([0.0, 0.5, 1.0, 1.5])

        # Use sparse_track which may have NaN values
        intervals = pm.gintervals("1", 0, 10000)

        result = pm.glookup(lookup_table, "sparse_track", breaks, intervals=intervals)

        # Should not crash; NaN in expression -> NaN in result
        assert result is not None

    def test_glookup_requires_lookup_table(self):
        """glookup requires a lookup_table argument."""
        breaks = np.array([0.0, 0.1, 0.2])
        intervals = pm.gintervals("1", 0, 100)

        with pytest.raises((ValueError, TypeError)):
            pm.glookup(None, "dense_track", breaks, intervals=intervals)

    def test_glookup_requires_intervals(self):
        """glookup requires intervals."""
        lookup_table = np.array([1, 2, 3])
        breaks = np.array([0.0, 0.1, 0.2, 0.3])

        with pytest.raises((ValueError, TypeError)):
            pm.glookup(lookup_table, "dense_track", breaks, intervals=None)


class TestGlookupIterator:
    """Test iterator parameter."""

    def test_glookup_with_iterator(self):
        """glookup with explicit iterator."""
        lookup_table = np.array([1, 2, 3])
        breaks = np.array([0.0, 0.15, 0.3, 0.45])
        intervals = pm.gintervals("1", 0, 500)

        result = pm.glookup(
            lookup_table, "dense_track", breaks,
            intervals=intervals, iterator=50
        )

        assert result is not None
        # With iterator=50, we should get more results than with default
        assert len(result) > 0


class TestGlookup2D:
    """Test 2D intervals/band behavior in Python fallback."""

    def test_glookup_2d_intervals_returns_2d_coordinates(self):
        """glookup on 2D intervals returns 2D coordinate columns."""
        lookup_table = np.array([1, 2, 3, 4, 5], dtype=float)
        breaks = np.linspace(0.0, 9000.0, 6)
        intervals = pm.gintervals_2d_all()

        result = pm.glookup(
            lookup_table,
            "rects_track", breaks,
            intervals=intervals,
        )

        assert result is not None
        expected_cols = {
            "chrom1", "start1", "end1", "chrom2", "start2", "end2", "intervalID", "value"
        }
        assert expected_cols.issubset(result.columns)
        assert len(result) > 0
        valid_values = result["value"].dropna().to_numpy(dtype=float)
        assert np.isin(valid_values, lookup_table).all()
        assert np.unique(valid_values).size > 1

    def test_glookup_2d_band_restricts_results(self):
        """band should reduce or keep (never increase) 2D lookup results."""
        lookup_table = np.array([1, 2, 3, 4, 5], dtype=float)
        breaks = np.linspace(0.0, 9000.0, 6)
        intervals = pm.gintervals_2d_all()

        full_result = pm.glookup(
            lookup_table,
            "rects_track", breaks,
            intervals=intervals,
        )
        band_result = pm.glookup(
            lookup_table,
            "rects_track", breaks,
            intervals=intervals,
            band=(-100, 100),
        )

        assert full_result is not None
        assert band_result is not None
        assert len(band_result) <= len(full_result)
        assert np.isin(
            band_result["value"].dropna().to_numpy(dtype=float),
            lookup_table,
        ).all()



"""Lookup and bin transform functions."""

from . import _shared
from ._shared import (
    CONFIG,
    _checkroot,
    _df2pymisha,
    _numpy,
    _pymisha,
    _pymisha2df,
)
from .expr import _find_vtracks_in_expr, _parse_expr_vars
from .extract import _is_2d_intervals, gextract
from .intervals import gintervals_2d_all, gintervals_all


def glookup(lookup_table, *args, intervals=None, include_lowest=False,
            force_binning=True, iterator=None, band=None, **kwargs):
    """
    Look up values in an N-dimensional lookup table indexed by track expressions.

    For each iterator interval, evaluates one or more track expressions and
    uses the resulting values to index into a lookup table. Returns the table
    value for each interval.

    Uses a memory-efficient C++ streaming implementation when expressions
    do not contain virtual tracks. Falls back to Python (memory-resident)
    when virtual tracks, a band filter, or 2D intervals are present.

    Parameters
    ----------
    lookup_table : numpy.ndarray
        N-dimensional lookup table. The shape must match the number of bins
        in each dimension. For 1D lookup, shape is ``(n_bins,)``. For
        multi-dimensional lookup, shape is ``(n_bins_1, n_bins_2, ...)``.
    *args : pairs of (str, array-like)
        Alternating track expressions and break arrays defining bins.
        Same format as ``gdist``.
        Example: ``glookup(table, "track1", breaks1, "track2", breaks2, ...)``.
    intervals : DataFrame or str
        Genomic scope for which the function is applied. Required.
    include_lowest : bool, default False
        If True, the lowest break value is included in the first bin.
        Example: ``breaks=[0, 0.2, 0.5]`` creates ``(0, 0.2], (0.2, 0.5]``.
        With ``include_lowest=True``: ``[0, 0.2], (0.2, 0.5]``.
    force_binning : bool, default True
        If True, clamp out-of-range values to the nearest bin instead of
        NaN. If False, out-of-range values produce NaN.
    iterator : int or str, optional
        Track expression iterator. If None, determined implicitly.
    band : tuple of (int, int), optional
        Diagonal band for 2D tracks. Triggers Python fallback path.

    Returns
    -------
    DataFrame or None
        Intervals with columns: ``chrom``, ``start``, ``end``, ``value``,
        ``intervalID``. Returns None if *intervals* is empty.

    See Also
    --------
    gdist : Compute distribution over binned track expressions.
    gtrack_lookup : Create a track from an N-dimensional lookup table.

    Examples
    --------
    >>> import pymisha as pm
    >>> import numpy as np
    >>> _ = pm.gdb_init_examples()

    One-dimensional lookup:

    >>> breaks = [0.1, 0.12, 0.14, 0.16, 0.18, 0.2]
    >>> result = pm.glookup([10, 20, 30, 40, 50], "dense_track", breaks,
    ...                     intervals=pm.gintervals("1", 0, 500))
    >>> "value" in result.columns
    True
    """
    if lookup_table is None:
        raise ValueError("lookup_table cannot be None")

    if intervals is None:
        raise ValueError("intervals cannot be None")

    # Parse arguments: (expr1, breaks1, expr2, breaks2, ...)
    if len(args) < 2:
        raise ValueError(
            "Usage: glookup(lookup_table, [expr, breaks]+, intervals=intervals, ...)"
        )

    if len(args) % 2 != 0:
        raise ValueError("glookup requires pairs of (expression, breaks) arguments")

    _checkroot()

    if len(intervals) == 0:
        return None

    lookup_table = _numpy.asarray(lookup_table, dtype=float)
    progress = kwargs.get("progress")
    progress_desc = kwargs.get("progress_desc", "glookup")

    exprs = []
    breaks_list = []

    for i in range(0, len(args), 2):
        expr = args[i]
        breaks = args[i + 1]

        if not isinstance(expr, str):
            raise TypeError(f"Expression at position {i} must be a string")

        breaks = _numpy.asarray(breaks, dtype=float)
        if breaks.ndim != 1:
            raise ValueError(f"Breaks at position {i+1} must be 1D array")
        if len(breaks) < 2:
            raise ValueError(f"Breaks at position {i+1} must have at least 2 elements")

        # Validate breaks are strictly increasing
        if not _numpy.all(_numpy.diff(breaks) > 0):
            raise ValueError(f"Breaks at position {i+1} must be strictly increasing")

        exprs.append(expr)
        breaks_list.append(breaks)

    # Validate lookup_table dimensions match number of expr-breaks pairs
    expected_dims = len(exprs)
    if lookup_table.ndim != expected_dims:
        raise ValueError(
            f"lookup_table has {lookup_table.ndim} dimensions but "
            f"{expected_dims} expr-breaks pairs were provided"
        )

    # Validate lookup_table shape matches number of bins
    n_bins = [len(b) - 1 for b in breaks_list]
    for i, (n, expected) in enumerate(zip(lookup_table.shape, n_bins, strict=False)):
        if n != expected:
            raise ValueError(
                f"lookup_table dimension {i} has size {n} but {expected} bins expected"
            )

    # Check if we need the Python path (vtracks, band, or 2D intervals)
    has_vtracks = any(_find_vtracks_in_expr(e) for e in exprs)
    use_python = has_vtracks or band is not None or _is_2d_intervals(intervals)

    if not use_python:
        # Use C++ streaming implementation
        config = dict(CONFIG)
        if progress is not None:
            config['progress'] = progress
        if progress_desc:
            config['progress_desc'] = progress_desc

        # Flatten lookup table in Fortran (column-major) order to match
        # BinsManager::vals2idx flat index convention
        flat_table = _numpy.ascontiguousarray(lookup_table.ravel(order='F'))

        result = _pymisha.pm_lookup(
            exprs,
            [b.tolist() for b in breaks_list],
            flat_table,
            _df2pymisha(intervals),
            iterator,
            include_lowest,
            force_binning,
            config,
        )

        if result is None:
            return None

        return _pymisha2df(result)

    # Fall back to Python implementation for virtual tracks / band / 2D
    return _glookup_python(
        lookup_table, exprs, breaks_list, intervals,
        include_lowest=include_lowest, force_binning=force_binning,
        iterator=iterator, band=band,
        progress=progress, progress_desc=progress_desc,
    )


def _glookup_python(lookup_table, exprs, breaks_list, intervals,
                    include_lowest=False, force_binning=True,
                    iterator=None, band=None,
                    progress=None, progress_desc=None):
    """Pure-Python fallback for glookup (used when vtracks, band, or 2D)."""
    all_values = []
    extract_result = None
    coord_cols = None

    for i, expr in enumerate(exprs):
        result = gextract(expr, intervals, iterator=iterator, band=band,
                         progress=progress, progress_desc=progress_desc)
        if result is None or len(result) == 0:
            return None

        coord_cols = {
            "chrom", "start", "end",
            "chrom1", "start1", "end1", "chrom2", "start2", "end2",
            "intervalID",
        }
        data_cols = [c for c in result.columns if c not in coord_cols]
        if not data_cols:
            return None

        values = result[data_cols[0]].to_numpy(dtype=float, copy=False)
        all_values.append(values)

        if i == 0:
            extract_result = result
            if {"chrom", "start", "end", "intervalID"}.issubset(result.columns):
                coord_cols = ["chrom", "start", "end", "intervalID"]
            elif {
                "chrom1", "start1", "end1", "chrom2", "start2", "end2", "intervalID"
            }.issubset(result.columns):
                coord_cols = ["chrom1", "start1", "end1", "chrom2", "start2", "end2", "intervalID"]
            else:
                raise ValueError("Unsupported extract result schema returned by gextract")
        else:
            if len(result) != len(extract_result):
                raise ValueError(
                    "glookup expression extraction produced mismatched row counts; "
                    "cannot align multi-expression lookup safely"
                )
            for c in coord_cols:
                if not _numpy.array_equal(
                    result[c].to_numpy(copy=False),
                    extract_result[c].to_numpy(copy=False),
                ):
                    raise ValueError(
                        f"glookup expression extraction produced misaligned rows (column '{c}')"
                    )

    n_values = len(all_values[0])

    # Bin each expression's values
    bin_indices = []
    for values, breaks in zip(all_values, breaks_list, strict=False):
        n_bin = len(breaks) - 1

        indices = _numpy.searchsorted(breaks, values, side='right') - 1

        if include_lowest:
            at_lowest = values == breaks[0]
            indices[at_lowest] = 0

        at_max = values == breaks[-1]
        indices[at_max] = n_bin - 1

        below_min = values < breaks[0]
        if not include_lowest:
            at_min = values == breaks[0]
            below_min = below_min | at_min

        above_max = values > breaks[-1]

        if force_binning:
            indices[below_min] = 0
            indices[above_max] = n_bin - 1
        else:
            indices[below_min] = -1
            indices[above_max] = -1

        nan_mask = _numpy.isnan(values)
        indices[nan_mask] = -1

        bin_indices.append(indices)

    # Look up values from table
    output_values = _numpy.full(n_values, _numpy.nan, dtype=float)

    valid = _numpy.ones(n_values, dtype=bool)
    for indices in bin_indices:
        valid &= (indices >= 0)

    if valid.any():
        if len(exprs) == 1:
            valid_indices = bin_indices[0][valid]
            output_values[valid] = lookup_table[valid_indices]
        else:
            valid_idx = [idx[valid] for idx in bin_indices]
            flat = _numpy.ravel_multi_index(valid_idx, lookup_table.shape)
            output_values[valid] = lookup_table.reshape(-1)[flat]

    result = extract_result[coord_cols].copy()
    result["value"] = output_values

    return result


def _resolve_lookup_dimensions(exprs, iterator):
    """Infer whether lookup output should be 1D or 2D from used physical tracks."""
    track_names = set(_pymisha.pm_track_names())
    if not track_names:
        return set()

    used_tracks = set()
    vtrack_names = _shared._VTRACKS

    for expr in exprs:
        _, parsed_tracks, _, _ = _parse_expr_vars(expr, track_names, vtrack_names)
        used_tracks.update(parsed_tracks)

    if isinstance(iterator, str) and iterator in track_names:
        used_tracks.add(iterator)

    if not used_tracks:
        return set()

    from .tracks import gtrack_info

    dims = set()
    for track in used_tracks:
        dims.add(int(gtrack_info(track).get("dimensions", 1)))
    return dims


def gtrack_lookup(track, description, lookup_table, *args, iterator=None,
                  include_lowest=False, force_binning=True, band=None):
    """
    Create a track from an N-dimensional lookup table.

    Evaluates track expressions genome-wide, looks up values in the table,
    and creates a new track from the results. Dense or sparse output is
    determined by the iterator type.

    This is the track-creation counterpart of :func:`glookup`, which returns
    values in-memory without creating a track.

    Parameters
    ----------
    track : str
        Name for the new track.
    description : str
        Track description.
    lookup_table : numpy.ndarray
        N-dimensional lookup table. Shape must match the number of bins
        defined by each ``(expr, breaks)`` pair. For 1D: ``(n_bins,)``.
        For multi-dimensional: ``(n_bins_1, n_bins_2, ...)``.
    *args : pairs of (str, array-like)
        Alternating track expressions and break arrays defining bins.
    iterator : int or str, optional
        Track expression iterator. Integer values create dense tracks
        with that bin size. Intervals-based iterators create sparse tracks.
    include_lowest : bool, default False
        If True, the lowest break value is included in the first bin.
    force_binning : bool, default True
        If True, clamp out-of-range values to the nearest bin. If False,
        out-of-range values produce NaN in the track.
    band : tuple of (int, int), optional
        Diagonal band for 2D tracks. Passed through to :func:`glookup`.

    Returns
    -------
    None

    See Also
    --------
    glookup : Look up values without creating a track.
    gtrack_create : Create a track from an expression.
    gdist : Compute distribution over binned track expressions.

    Examples
    --------
    >>> import pymisha as pm
    >>> import numpy as np
    >>> _ = pm.gdb_init_examples()

    Create a dense track from 1D lookup:

    >>> pm.gtrack_lookup("my_track", "lookup track",  # doctest: +SKIP
    ...     np.array([10.0, 20.0, 30.0, 40.0]),
    ...     "dense_track", [0, 0.05, 0.1, 0.15, 0.2],
    ...     iterator=100)
    """
    if lookup_table is None:
        raise ValueError("lookup_table cannot be None")

    # Parse arguments: (expr1, breaks1, expr2, breaks2, ...)
    if len(args) < 2:
        raise ValueError(
            "Usage: gtrack_lookup(track, description, lookup_table, "
            "[expr, breaks]+, iterator=..., ...)"
        )

    if len(args) % 2 != 0:
        raise ValueError("gtrack_lookup requires pairs of (expression, breaks) arguments")

    _checkroot()

    lookup_table = _numpy.asarray(lookup_table, dtype=float)

    exprs = []
    breaks_list = []

    for i in range(0, len(args), 2):
        expr = args[i]
        breaks = args[i + 1]

        if not isinstance(expr, str):
            raise TypeError(f"Expression at position {i} must be a string")

        breaks = _numpy.asarray(breaks, dtype=float)
        if breaks.ndim != 1:
            raise ValueError(f"Breaks at position {i+1} must be 1D array")
        if len(breaks) < 2:
            raise ValueError(f"Breaks at position {i+1} must have at least 2 elements")
        if not _numpy.all(_numpy.diff(breaks) > 0):
            raise ValueError(f"Breaks at position {i+1} must be strictly increasing")

        exprs.append(expr)
        breaks_list.append(breaks)

    # Validate lookup_table dimensions
    expected_dims = len(exprs)
    if lookup_table.ndim != expected_dims and not (lookup_table.ndim == 1 and expected_dims == 1):
        raise ValueError(
            f"lookup_table has {lookup_table.ndim} dimensions but "
                f"{expected_dims} expr-breaks pairs were provided"
            )

    n_bins = [len(b) - 1 for b in breaks_list]
    for i, (n, expected) in enumerate(zip(lookup_table.shape, n_bins, strict=False)):
        if n != expected:
            raise ValueError(
                f"lookup_table dimension {i} has size {n} but {expected} bins expected"
            )

    # Lazy imports to avoid circular dependency
    import shutil

    from .tracks import (
        _ensure_track_absent,
        _track_dir_for_create,
        _validate_track_name,
        gtrack_create_dense,
        gtrack_create_sparse,
    )

    _validate_track_name(track)
    _ensure_track_absent(track)

    dims = _resolve_lookup_dimensions(exprs, iterator)
    if 1 in dims and 2 in dims:
        raise ValueError("Cannot mix 1D and 2D tracks in gtrack_lookup expressions/iterator")

    use_2d_scope = (band is not None) or (2 in dims)
    scope_intervals = gintervals_2d_all() if use_2d_scope else gintervals_all()

    # Compute lookup values using glookup over the whole genome
    result = glookup(
        lookup_table, *args,
        intervals=scope_intervals,
        include_lowest=include_lowest,
        force_binning=force_binning,
        iterator=iterator,
        band=band,
    )

    if result is None or len(result) == 0:
        raise ValueError("Lookup produced no values; cannot create track")

    is_2d_result = {
        "chrom1", "start1", "end1", "chrom2", "start2", "end2"
    }.issubset(result.columns)
    if is_2d_result:
        intervals_df = result[["chrom1", "start1", "end1", "chrom2", "start2", "end2"]].copy()
    else:
        intervals_df = result[["chrom", "start", "end"]].copy()
    values = result["value"].to_numpy(dtype=float)

    # Determine track type from iterator for 1D outputs.
    # If iterator is a positive integer, create dense track; otherwise sparse.
    is_dense = (
        (not is_2d_result)
        and isinstance(iterator, (int, float, _numpy.integer, _numpy.floating))
        and int(iterator) > 0
    )

    track_dir = _track_dir_for_create(track)
    created_new = not track_dir.exists()
    try:
        if is_2d_result:
            from .tracks import gtrack_2d_create
            gtrack_2d_create(track, description, intervals_df, values)
        elif is_dense:
            binsize = int(iterator)
            gtrack_create_dense(track, description, intervals_df, values,
                                binsize, defval=_numpy.nan)
        else:
            gtrack_create_sparse(track, description, intervals_df, values)

        # Build created.by string matching R convention
        exprs_str = ", ".join(
            f'"{e}", c({", ".join(str(v) for v in b)})'
            for e, b in zip(exprs, breaks_list, strict=False)
        )
        created_by = (
            f'gtrack.lookup("{track}", description, lookup_table, '
            f'{exprs_str}, iterator={iterator!r})'
        )
        # Bypass readonly check for internal track creation
        from .tracks import _load_track_attributes, _save_track_attributes
        attrs = _load_track_attributes(track)
        attrs["created.by"] = created_by
        _save_track_attributes(track, attrs)

    except Exception:
        if created_new and track_dir.exists():
            shutil.rmtree(track_dir, ignore_errors=True)
            _pymisha.pm_dbreload()
        raise



import conftest
import numpy as np
import pytest

import pymisha as pm


def test_gquantiles_dense_track_matches_numpy():
    intervals = pm.gintervals_all()
    percentiles = [0.1, 0.5, 0.9]

    values = conftest.extract_values("dense_track", intervals)
    expected = np.nanquantile(values, percentiles)

    result = pm.gquantiles("dense_track", percentiles, intervals)
    np.testing.assert_allclose(result.to_numpy(), expected, rtol=1e-12, atol=1e-12)


def test_gquantiles_accepts_scalar_percentile():
    intervals = pm.gintervals_all()
    values = conftest.extract_values("dense_track", intervals)
    expected = np.nanquantile(values, [0.5])[0]

    result = pm.gquantiles("dense_track", 0.5, intervals)
    assert result.iloc[0] == pytest.approx(expected)


def test_gquantiles_rejects_out_of_range_percentiles():
    intervals = pm.gintervals_all()
    with pytest.raises(ValueError, match=r"percentiles must be within \[0, 1\]"):
        pm.gquantiles("dense_track", [-0.1, 1.1], intervals)


def test_gquantiles_sparse_track_matches_numpy():
    intervals = pm.gintervals_all()
    percentiles = [0.25, 0.5, 0.75]

    values = conftest.extract_values("sparse_track", intervals)
    expected = np.nanquantile(values, percentiles)

    result = pm.gquantiles("sparse_track", percentiles, intervals)
    np.testing.assert_allclose(result.to_numpy(), expected, rtol=1e-12, atol=1e-12)



"""Expression parsing helpers."""

import re

from . import _shared

_IDENTIFIER_RE = re.compile(r"^[A-Za-z_][A-Za-z0-9_.]*$")
_IDENTIFIER_TOKEN_RE = re.compile(r"[A-Za-z_][A-Za-z0-9_.]*")
_TOKEN_RE = re.compile(
    r"""
    \s+|
    [A-Za-z_][A-Za-z0-9_.]*|
    \d+(?:\.\d*)?(?:[eE][+-]?\d+)?|
    \.\d+(?:[eE][+-]?\d+)?|
    ==|!=|<=|>=|//|\*\*|
    \S
    """,
    re.VERBOSE,
)


def _expr_safe_name(name):
    """Return a collision-proof identifier used in expression eval namespaces."""
    return "__pmv_" + name.encode("utf-8").hex()


def _register_expr_name(name, track_names, vtrack_names, used_tracks, used_vtracks, var_map):
    safe = _expr_safe_name(name)
    if name in track_names:
        used_tracks.add(name)
    if name in vtrack_names:
        used_vtracks.add(name)
    var_map[safe] = name
    return safe


def _replace_identifier_token(token, track_names, vtrack_names, used_tracks, used_vtracks, var_map):
    if token in track_names or token in vtrack_names:
        return _register_expr_name(
            token, track_names, vtrack_names, used_tracks, used_vtracks, var_map
        )

    # Tokenizer keeps dotted names together. If the full token is unknown,
    # resolve the longest known dotted prefix and preserve the suffix.
    if "." in token:
        parts = token.split(".")
        for i in range(len(parts) - 1, 0, -1):
            prefix = ".".join(parts[:i])
            if prefix in track_names or prefix in vtrack_names:
                safe = _register_expr_name(
                    prefix, track_names, vtrack_names, used_tracks, used_vtracks, var_map
                )
                return safe + token[len(prefix):]

    return None


def _find_vtracks_in_expr(expr):
    """Find virtual track names used in an expression."""
    if not _shared._VTRACKS:
        return []
    known_vtracks = set(_shared._VTRACKS)
    matched = set()

    for token in set(_IDENTIFIER_TOKEN_RE.findall(expr)):
        if token in known_vtracks:
            matched.add(token)
            continue
        if "." in token:
            parts = token.split(".")
            for i in range(len(parts) - 1, 0, -1):
                prefix = ".".join(parts[:i])
                if prefix in known_vtracks:
                    matched.add(prefix)
                    break

    return [name for name in _shared._VTRACKS if name in matched]


def _parse_expr_vars(expr, track_names, vtrack_names):
    """
    Parse an expression and replace track/vtrack names with safe Python identifiers.

    Returns:
        new_expr, used_tracks, used_vtracks, var_map
    """
    tokens = _TOKEN_RE.findall(expr)

    used_tracks = set()
    used_vtracks = set()
    var_map = {}
    out = []

    for token in tokens:
        if _IDENTIFIER_RE.fullmatch(token):
            replaced = _replace_identifier_token(
                token, track_names, vtrack_names, used_tracks, used_vtracks, var_map
            )
            if replaced is not None:
                out.append(replaced)
                continue
        out.append(token)

    return ''.join(out), used_tracks, used_vtracks, var_map



"""Tests for gtrack_lookup: track creation via lookup table."""

import contextlib

import numpy as np
import pytest

import pymisha as pm


@pytest.fixture(autouse=True)
def setup_db():
    """Initialize example database for each test."""
    pm.gdb_init_examples()
    yield
    # Clean up any test tracks
    for t in pm.gtrack_ls():
        if t.startswith("test_lookup_"):
            with contextlib.suppress(Exception):
                pm.gtrack_rm(t, force=True)


class TestGtrackLookupBasic:
    """Basic gtrack_lookup functionality."""

    def test_creates_track_1d(self):
        """gtrack_lookup creates a 1D track from a lookup table."""
        breaks = [0, 0.05, 0.1, 0.15, 0.2]
        table = np.array([10.0, 20.0, 30.0, 40.0])

        pm.gtrack_lookup(
            "test_lookup_1d", "", table,
            "dense_track", breaks,
            iterator=100,
        )

        assert pm.gtrack_exists("test_lookup_1d")
        info = pm.gtrack_info("test_lookup_1d")
        assert info is not None

    def test_track_values_match_glookup(self):
        """Values in created track should match glookup output."""
        breaks = [0, 0.05, 0.1, 0.15, 0.2]
        table = np.array([10.0, 20.0, 30.0, 40.0])
        intervals = pm.gintervals("1", 0, 5000)

        pm.gtrack_lookup(
            "test_lookup_match", "", table,
            "dense_track", breaks,
            iterator=100,
        )

        # Extract values from the created track
        track_vals = pm.gextract("test_lookup_match", intervals, iterator=100)

        # Get values from glookup
        lookup_vals = pm.glookup(table, "dense_track", breaks,
                                 intervals=intervals, iterator=100)

        if track_vals is not None and lookup_vals is not None:
            # Compare the value columns
            track_data = track_vals.iloc[:, 3:-1].to_numpy(dtype=float).ravel()
            lookup_data = lookup_vals["value"].to_numpy(dtype=float)
            # They should match (NaN-equal comparison)
            np.testing.assert_array_equal(
                np.where(np.isnan(track_data), -999, track_data),
                np.where(np.isnan(lookup_data), -999, lookup_data),
            )

    def test_creates_track_2d_lookup(self):
        """gtrack_lookup with 2 expression-breaks pairs creates a track."""
        breaks1 = [0, 0.05, 0.1, 0.15, 0.2]
        breaks2 = [0, 0.1, 0.2]
        table = np.arange(1, 9, dtype=float).reshape((4, 2))

        pm.gtrack_lookup(
            "test_lookup_2d", "", table,
            "dense_track", breaks1,
            "dense_track", breaks2,
            iterator=100,
        )

        assert pm.gtrack_exists("test_lookup_2d")

    def test_include_lowest(self):
        """gtrack_lookup respects include_lowest parameter."""
        breaks = [0, 0.1, 0.2, 0.5]
        table = np.array([100.0, 200.0, 300.0])
        intervals = pm.gintervals("1", 0, 5000)

        pm.gtrack_lookup(
            "test_lookup_il_false", "", table,
            "dense_track", breaks,
            iterator=100,
            include_lowest=False,
        )
        pm.gtrack_lookup(
            "test_lookup_il_true", "", table,
            "dense_track", breaks,
            iterator=100,
            include_lowest=True,
        )

        vals_false = pm.gextract("test_lookup_il_false", intervals, iterator=100)
        vals_true = pm.gextract("test_lookup_il_true", intervals, iterator=100)

        if vals_false is not None and vals_true is not None:
            # With include_lowest=True, fewer NaN values (values at breaks[0]
            # get mapped to first bin instead of NaN)
            n_nan_false = vals_false.iloc[:, 3:-1].isna().sum().sum()
            n_nan_true = vals_true.iloc[:, 3:-1].isna().sum().sum()
            assert n_nan_true <= n_nan_false

    def test_force_binning_false(self):
        """gtrack_lookup with force_binning=False produces NaN for out-of-range."""
        breaks = [0.05, 0.1, 0.15]
        table = np.array([100.0, 200.0])
        intervals = pm.gintervals("1", 0, 5000)

        pm.gtrack_lookup(
            "test_lookup_nofb", "", table,
            "dense_track", breaks,
            iterator=100,
            force_binning=False,
        )

        vals = pm.gextract("test_lookup_nofb", intervals, iterator=100)
        if vals is not None:
            data = vals.iloc[:, 3:-1].to_numpy(dtype=float).ravel()
            # Some values should be NaN (out of range)
            assert np.any(np.isnan(data))
            # Valid values should only be from the table
            valid = data[~np.isnan(data)]
            assert all(v in table for v in valid)

    def test_force_binning_true(self):
        """gtrack_lookup with force_binning=True clamps out-of-range."""
        breaks = [0.05, 0.1, 0.15]
        table = np.array([100.0, 200.0])
        intervals = pm.gintervals("1", 0, 5000)

        pm.gtrack_lookup(
            "test_lookup_fb", "", table,
            "dense_track", breaks,
            iterator=100,
            force_binning=True,
        )

        vals = pm.gextract("test_lookup_fb", intervals, iterator=100)
        if vals is not None:
            data = vals.iloc[:, 3:-1].to_numpy(dtype=float).ravel()
            # With force_binning, only NaN should come from NaN source values
            valid = data[~np.isnan(data)]
            assert all(v in table for v in valid)

    def test_creates_2d_track_from_2d_expression(self):
        """2D source expressions should produce a 2D output track."""
        breaks = np.linspace(0.0, 9000.0, 6)
        table = np.array([1.0, 2.0, 3.0, 4.0, 5.0])

        pm.gtrack_lookup(
            "test_lookup_2d_rects", "", table,
            "rects_track", breaks,
        )

        assert pm.gtrack_exists("test_lookup_2d_rects")
        info = pm.gtrack_info("test_lookup_2d_rects")
        assert info.get("dimensions") == 2

        extracted = pm.gextract("test_lookup_2d_rects", pm.gintervals_2d_all())
        assert extracted is not None
        assert len(extracted) > 0

    def test_band_creates_filtered_2d_track(self):
        """band parameter should be accepted and applied for 2D lookup tracks."""
        breaks = np.linspace(0.0, 9000.0, 6)
        table = np.array([1.0, 2.0, 3.0, 4.0, 5.0])

        pm.gtrack_lookup(
            "test_lookup_2d_full", "", table,
            "rects_track", breaks,
        )
        pm.gtrack_lookup(
            "test_lookup_2d_band", "", table,
            "rects_track", breaks,
            band=(-100, 100),
        )

        full_res = pm.gextract("test_lookup_2d_full", pm.gintervals_2d_all())
        band_res = pm.gextract("test_lookup_2d_band", pm.gintervals_2d_all())

        assert full_res is not None
        assert band_res is not None
        assert len(band_res) <= len(full_res)


class TestGtrackLookupAttributes:
    """Test that gtrack_lookup sets proper track attributes."""

    def test_sets_description(self):
        """Created track should have the given description."""
        breaks = [0, 0.1, 0.2]
        table = np.array([1.0, 2.0])

        pm.gtrack_lookup(
            "test_lookup_desc", "my test description", table,
            "dense_track", breaks,
            iterator=100,
        )

        desc = pm.gtrack_attr_get("test_lookup_desc", "description")
        assert desc == "my test description"

    def test_sets_created_by(self):
        """Created track should have created.by attribute."""
        breaks = [0, 0.1, 0.2]
        table = np.array([1.0, 2.0])

        pm.gtrack_lookup(
            "test_lookup_cb", "", table,
            "dense_track", breaks,
            iterator=100,
        )

        created_by = pm.gtrack_attr_get("test_lookup_cb", "created.by")
        assert created_by is not None
        assert "gtrack.lookup" in created_by


class TestGtrackLookupValidation:
    """Test argument validation."""

    def test_rejects_existing_track(self):
        """Should reject creating a track that already exists."""
        breaks = [0, 0.1, 0.2]
        table = np.array([1.0, 2.0])

        with pytest.raises(Exception):
            pm.gtrack_lookup(
                "dense_track", "", table,
                "dense_track", breaks,
                iterator=100,
            )

    def test_rejects_bad_table_dimensions(self):
        """Should reject lookup table with wrong dimensions."""
        breaks1 = [0, 0.1, 0.2]
        breaks2 = [0, 0.5, 1.0]
        table = np.array([1.0, 2.0])  # 1D table for 2D lookup

        with pytest.raises(ValueError):
            pm.gtrack_lookup(
                "test_lookup_bad_dim", "", table,
                "dense_track", breaks1,
                "dense_track", breaks2,
                iterator=100,
            )

    def test_rejects_bad_table_shape(self):
        """Should reject lookup table with wrong shape."""
        breaks = [0, 0.1, 0.2, 0.3]  # 3 bins
        table = np.array([1.0, 2.0])  # 2 elements, need 3

        with pytest.raises(ValueError):
            pm.gtrack_lookup(
                "test_lookup_bad_shape", "", table,
                "dense_track", breaks,
                iterator=100,
            )

    def test_rejects_no_args(self):
        """Should reject missing expression arguments."""
        table = np.array([1.0])

        with pytest.raises((ValueError, TypeError)):
            pm.gtrack_lookup("test_lookup_noarg", "", table)

    def test_rejects_odd_args(self):
        """Should reject odd number of expression arguments."""
        table = np.array([1.0, 2.0])

        with pytest.raises(ValueError):
            pm.gtrack_lookup(
                "test_lookup_odd", "", table,
                "dense_track",  # missing breaks
            )


class TestGtrackLookupRParity:
    """R misha parity tests ported from test-gtrack.lookup.R."""

    def test_lookup_default_binning_1d(self):
        """Port of: lookup and extract with default binning (1D dense)."""
        breaks = np.linspace(0, 0.3, 6).tolist()
        table = np.array([10.0, 20.0, 30.0, 40.0, 50.0])
        trackname = "test_lookup_r_default"
        try:
            pm.gtrack_lookup(trackname, "", table,
                             "dense_track", breaks, iterator=50)
            r = pm.gextract(trackname, pm.gintervals([1, 2]))
            assert r is not None
            assert len(r) == 16000
            vals = r[trackname].values
            non_nan = vals[~np.isnan(vals)]
            # All non-NaN values should come from the table
            assert set(non_nan).issubset(set(table))
        finally:
            with contextlib.suppress(Exception):
                pm.gtrack_rm(trackname, force=True)

    def test_lookup_no_force_binning(self):
        """Port of: lookup and extract without force binning."""
        breaks = np.linspace(0.1, 0.2, 6).tolist()
        table = np.array([10.0, 20.0, 30.0, 40.0, 50.0])
        trackname = "test_lookup_r_nofb"
        try:
            pm.gtrack_lookup(trackname, "", table,
                             "dense_track", breaks,
                             force_binning=False, iterator=50)
            r = pm.gextract(trackname, pm.gintervals([1, 2]))
            assert r is not None
            vals = r[trackname].values
            non_nan = vals[~np.isnan(vals)]
            nan_count = np.isnan(vals).sum()
            # Without force_binning, values outside range should be NaN
            assert nan_count > 0
            assert set(non_nan).issubset(set(table))
        finally:
            with contextlib.suppress(Exception):
                pm.gtrack_rm(trackname, force=True)

    def test_lookup_2d_dense_dimensions(self):
        """Port of: lookup with 2D expressions creates correct track."""
        m1 = np.arange(1, 16, dtype=float).reshape((5, 3), order='F')
        breaks1 = np.linspace(0.1, 0.2, 6).tolist()
        breaks2 = np.linspace(0.1, 0.5, 4).tolist()
        trackname = "test_lookup_r_2d"
        try:
            pm.gtrack_lookup(trackname, "", m1,
                             "dense_track", breaks1,
                             "dense_track", breaks2,
                             iterator=50)
            assert pm.gtrack_exists(trackname)
            r = pm.gextract(trackname, pm.gintervals([1, 2]))
            assert r is not None
            assert len(r) > 0
            vals = r[trackname].values
            non_nan = vals[~np.isnan(vals)]
            # Values should come from the matrix entries
            assert len(non_nan) > 0
            assert all(1 <= v <= 15 for v in non_nan)
        finally:
            with contextlib.suppress(Exception):
                pm.gtrack_rm(trackname, force=True)
